"""Utilities for computing metrics.

NOTE that only the widely used metrics are implemented here,
challenge (e.g. CinC, CPSC series) specific metrics are not included.

"""

import warnings
from numbers import Real
from typing import Dict, Optional, Sequence, Tuple, Union

import einops
import numpy as np
import torch
from torch import Tensor

from .misc import add_docstring
from .utils_data import ECGWaveForm, ECGWaveFormNames, masks_to_waveforms
from ..cfg import DEFAULTS


__all__ = [
    "top_n_accuracy",
    "confusion_matrix",
    "ovr_confusion_matrix",
    "QRS_score",
    "metrics_from_confusion_matrix",
    "compute_wave_delineation_metrics",
]


def top_n_accuracy(
    labels: Union[np.ndarray, Tensor],
    outputs: Union[np.ndarray, Tensor],
    n: Union[int, Sequence[int]] = 1,
) -> Union[float, Dict[str, float]]:
    """Compute top n accuracy.

    Parameters
    ----------
    labels : numpy.ndarray or torch.Tensor
        Labels of class indices,
        of shape ``(batch_size,)`` or ``(batch_size, d_1, ..., d_m)``.
    outputs : numpy.ndarray or torch.Tensor
        Predicted probabilities, of shape ``(batch_size, num_classes)``
        or ``(batch_size, d_1, ..., d_m, num_classes)``
        or ``(batch_size, num_classes, d_1, ..., d_m)``.
    n : int or List[int]
        Top n to be considered.

    Returns
    -------
    acc : float or dict of float
        Top n accuracy.

    Examples
    --------
    >>> from torch_ecg.cfg import DEFAULTS
    >>> labels, outputs = DEFAULTS.RNG_randint(0, 9, (100)), DEFAULTS.RNG.uniform(0, 1, (100, 10))  # 100 samples, 10 classes
    >>> top_n_accuracy(labels, outputs, 3)
    0.32
    >>> top_n_accuracy(labels, outputs, [1,3,5])
    {'top_1_acc': 0.12, 'top_3_acc': 0.32, 'top_5_acc': 0.52}

    """
    assert (
        outputs.shape[0] == labels.shape[0]
    ), "outputs and labels must have the same batch size"
    labels, outputs = torch.as_tensor(labels), torch.as_tensor(outputs)
    batch_size, n_classes, *extra_dims = outputs.shape
    if isinstance(n, int):
        ln = [n]
    else:
        ln = n
    acc = {}
    for _n in ln:
        key = f"top_{_n}_acc"
        _, indices = torch.topk(
            outputs, _n, dim=1
        )  # of shape (batch_size, n) or (batch_size, n, d_1, ..., d_n)
        pattern = " ".join([f"d_{i+1}" for i in range(len(extra_dims))])
        pattern = f"batch_size {pattern} -> batch_size n {pattern}"
        correct = torch.sum(indices == einops.repeat(labels, pattern, n=_n))
        acc[key] = correct.item() / outputs.shape[0]
        for d in extra_dims:
            acc[key] = acc[key] / d
    if len(ln) == 1:
        return acc[key]
    return acc


def confusion_matrix(
    labels: Union[np.ndarray, Tensor],
    outputs: Union[np.ndarray, Tensor],
    num_classes: Optional[int] = None,
) -> np.ndarray:
    """Compute a binary confusion matrix

    The columns are ground truth labels and rows are predicted labels.

    Parameters
    ----------
    labels : numpy.ndarray or torch.Tensor
        Binary labels, of shape ``(n_samples, n_classes)``,
        or indices of each label class, of shape ``(n_samples,)``.
    outputs : numpy.ndarray or torch.Tensor
        Binary outputs, of shape ``(n_samples, n_classes)``,
        or indices of each class predicted, of shape ``(n_samples,)``.
    num_classes : int, optional
        Number of classes.
        If `labels` and `outputs` are both of shape ``(n_samples,)``,
        then `num_classes` must be specified.

    Returns
    -------
    cm : numpy.ndarray
        Confusion matrix, of shape ``(n_classes, n_classes)``.

    """
    labels, outputs = cls_to_bin(labels, outputs, num_classes)
    assert np.shape(labels) == np.shape(
        outputs
    ), "labels and outputs must have the same shape"
    assert all(value in (0, 1) for value in np.unique(labels)), "labels must be binary"
    assert all(
        value in (0, 1) for value in np.unique(outputs)
    ), "outputs must be binary"

    num_samples, num_classes = np.shape(labels)

    cm = np.zeros((num_classes, num_classes))
    for k in range(num_samples):
        i = np.argmax(outputs[k, :])
        j = np.argmax(labels[k, :])
        cm[i, j] += 1

    return cm


def one_vs_rest_confusion_matrix(
    labels: Union[np.ndarray, Tensor],
    outputs: Union[np.ndarray, Tensor],
    num_classes: Optional[int] = None,
) -> np.ndarray:
    """Compute binary one-vs-rest confusion matrices.

    Columns are ground truth labels and rows are predicted labels.

    Parameters
    ----------
    labels : numpy.ndarray or torch.Tensor
        Binary labels, of shape ``(n_samples, n_classes)``,
        or indices of each label class, of shape ``(n_samples,)``.
    outputs : numpy.ndarray or torch.Tensor
        Binary outputs, of shape ``(n_samples, n_classes)``,
        or indices of each class predicted, of shape ``(n_samples,)``.
    num_classes : int, optional
        number of classes.
        If `labels` and `outputs` are both of shape ``(n_samples,)``,
        then `num_classes` must be specified.

    Returns
    -------
    ovr_cm : numpy.ndarray
        One-vs-rest confusion matrix, of shape ``(n_classes, 2, 2)``.

    """
    labels, outputs = cls_to_bin(labels, outputs, num_classes)
    assert np.shape(labels) == np.shape(
        outputs
    ), "labels and outputs must have the same shape"
    assert all(value in (0, 1) for value in np.unique(labels)), "labels must be binary"
    assert all(
        value in (0, 1) for value in np.unique(outputs)
    ), "outputs must be binary"

    num_samples, num_classes = np.shape(labels)

    ovr_cm = np.zeros((num_classes, 2, 2))
    for i in range(num_samples):
        for j in range(num_classes):
            if labels[i, j] == 1 and outputs[i, j] == 1:  # TP
                ovr_cm[j, 0, 0] += 1
            elif labels[i, j] == 0 and outputs[i, j] == 1:  # FP
                ovr_cm[j, 0, 1] += 1
            elif labels[i, j] == 1 and outputs[i, j] == 0:  # FN
                ovr_cm[j, 1, 0] += 1
            elif labels[i, j] == 0 and outputs[i, j] == 0:  # TN
                ovr_cm[j, 1, 1] += 1

    return ovr_cm


# alias
ovr_confusion_matrix = one_vs_rest_confusion_matrix


_METRICS_FROM_CONFUSION_MATRIX_PARAMS = """
    Compute macro {metric}, and {metrics} for each class.

    Parameters
    ----------
    labels : numpy.ndarray or torch.Tensor
        Binary labels, of shape ``(n_samples, n_classes)``,
        or indices of each label class, of shape ``(n_samples,)``.
    outputs : numpy.ndarray or torch.Tensor
        Probability outputs, of shape ``(n_samples, n_classes)``,
        or binary outputs, of shape ``(n_samples, n_classes)``,
        or indices of each class predicted, of shape ``(n_samples,)``.
    num_classes : int, optional
        Number of classes.
        If `labels` and `outputs` are both of shape ``(n_samples,)``,
        then `num_classes` must be specified.
    weights : numpy.ndarray or torch.Tensor, optional
        Weights for each class, of shape ``(n_classes,)``,
        used to compute macro {metric}.
    thr : float, default: 0.5
        Threshold for binary classification,
        valid only if `outputs` is of shape ``(n_samples, n_classes)``.
"""


@add_docstring(
    _METRICS_FROM_CONFUSION_MATRIX_PARAMS.format(metric="metrics", metrics="metrics"),
    "prepend",
)
def metrics_from_confusion_matrix(
    labels: Union[np.ndarray, Tensor],
    outputs: Union[np.ndarray, Tensor],
    num_classes: Optional[int] = None,
    weights: Optional[Union[np.ndarray, Tensor]] = None,
    thr: float = 0.5,
) -> Dict[str, Union[float, np.ndarray]]:
    """
    Returns
    -------
    metrics : dict
        Metrics computed from the one-vs-rest confusion matrix.

    Examples
    --------
    >>> from torch_ecg.cfg import DEFAULTS
    >>> # binary labels (100 samples, 10 classes, multi-label)
    >>> labels = DEFAULTS.RNG_randint(0, 1, (100, 10))
    >>> # probability outputs (100 samples, 10 classes, multi-label)
    >>> outputs = DEFAULTS.RNG.random((100, 10))
    >>> metrics = metrics_from_confusion_matrix(labels, outputs)
    >>> # binarize outputs (100 samples, 10 classes, multi-label)
    >>> outputs = DEFAULTS.RNG_randint(0, 1, (100, 10))
    >>> # would raise
    >>> # RuntimeWarning: `outputs` is probably binary, AUC may be incorrect
    >>> metrics = metrics_from_confusion_matrix(labels, outputs)
    >>> # categorical outputs (100 samples, 10 classes)
    >>> outputs = DEFAULTS.RNG_randint(0, 9, (100,))
    >>> # would raise
    >>> # RuntimeWarning: `outputs` is probably binary, AUC may be incorrect
    >>> metrics = metrics_from_confusion_matrix(labels, outputs)

    """
    outputs_ndim = np.ndim(outputs)
    labels, outputs = cls_to_bin(labels, outputs, num_classes)
    num_samples, num_classes = np.shape(labels)

    # probability outputs to binary outputs
    bin_outputs = np.zeros_like(outputs, dtype=int)
    bin_outputs[outputs >= thr] = 1
    bin_outputs[outputs < thr] = 0
    if np.unique(outputs).size == 2:
        warnings.warn(
            "`outputs` is probably binary, AUC may be incorrect", RuntimeWarning
        )

    ovr_cm = ovr_confusion_matrix(labels, bin_outputs)

    # sens: sensitivity, recall, hit rate, or true positive rate
    # spec: specificity, selectivity or true negative rate
    # prec: precision or positive predictive value
    # npv: negative predictive value
    # jac: jaccard index, threat score, or critical success index
    # acc: accuracy
    # phi: phi coefficient, or matthews correlation coefficient
    # NOTE: never use repeat here, because it will cause bugs
    # sens, spec, prec, npv, jac, acc, phi = list(repeat(np.zeros(num_classes), 7))
    sens, spec, prec, npv, jac, acc, phi = [np.zeros(num_classes) for _ in range(7)]
    auroc = np.zeros(
        num_classes
    )  # area under the receiver-operater characteristic curve (ROC AUC)
    auprc = np.zeros(num_classes)  # area under the precision-recall curve
    for k in range(num_classes):
        tp, fp, fn, tn = (
            ovr_cm[k, 0, 0],
            ovr_cm[k, 0, 1],
            ovr_cm[k, 1, 0],
            ovr_cm[k, 1, 1],
        )
        if tp + fn > 0:
            sens[k] = tp / (tp + fn)
        else:
            sens[k] = float("nan")
        if tp + fp > 0:
            prec[k] = tp / (tp + fp)
        else:
            prec[k] = float("nan")
        if tn + fp > 0:
            spec[k] = tn / (tn + fp)
        else:
            spec[k] = float("nan")
        if tn + fn > 0:
            npv[k] = tn / (tn + fn)
        else:
            npv[k] = float("nan")
        if tp + fn + fp > 0:
            jac[k] = tp / (tp + fn + fp)
        else:
            jac[k] = float("nan")
        acc[k] = (tp + tn) / num_samples
        phi[k] = (tp * tn - fp * fn) / np.sqrt(
            (tp + fp) * (tp + fn) * (tn + fp) * (tn + fn)
        )

        if outputs_ndim == 1:
            auroc[k] = np.nan
            auprc[k] = np.nan
            continue
        thresholds = np.unique(outputs[:, k])
        thresholds = np.append(thresholds, thresholds[-1] + 1)
        thresholds = thresholds[::-1]
        num_thresholds = len(thresholds)

        # Initialize the TPs, FPs, FNs, and TNs.
        tp = np.zeros(num_thresholds)
        fp = np.zeros(num_thresholds)
        fn = np.zeros(num_thresholds)
        tn = np.zeros(num_thresholds)
        fn[0] = np.sum(labels[:, k] == 1)
        tn[0] = np.sum(labels[:, k] == 0)

        # Find the indices that result in sorted output values.
        idx = np.argsort(outputs[:, k])[::-1]

        # Compute the TPs, FPs, FNs, and TNs for class k across thresholds.
        i = 0
        for j in range(1, num_thresholds):
            # Initialize TPs, FPs, FNs, and TNs using values at previous threshold.
            tp[j] = tp[j - 1]
            fp[j] = fp[j - 1]
            fn[j] = fn[j - 1]
            tn[j] = tn[j - 1]

            # Update the TPs, FPs, FNs, and TNs at i-th output value.
            while i < num_samples and outputs[idx[i], k] >= thresholds[j]:
                if labels[idx[i], k]:
                    tp[j] += 1
                    fn[j] -= 1
                else:
                    fp[j] += 1
                    tn[j] -= 1
                i += 1

        # Summarize the TPs, FPs, FNs, and TNs for class k.
        tpr_ = np.zeros(num_thresholds)
        tnr_ = np.zeros(num_thresholds)
        ppv_ = np.zeros(num_thresholds)
        for j in range(num_thresholds):
            if tp[j] + fn[j]:
                tpr_[j] = float(tp[j]) / float(tp[j] + fn[j])
            else:
                tpr_[j] = float("nan")
            if fp[j] + tn[j]:
                tnr_[j] = float(tn[j]) / float(fp[j] + tn[j])
            else:
                tnr_[j] = float("nan")
            if tp[j] + fp[j]:
                ppv_[j] = float(tp[j]) / float(tp[j] + fp[j])
            else:
                ppv_[j] = float("nan")

        # Compute AUROC as the area under a piecewise linear function with TPR/
        # sensitivity (x-axis) and TNR/specificity (y-axis) and AUPRC as the area
        # under a piecewise constant with TPR/recall (x-axis) and PPV/precision
        # (y-axis) for class k.
        for j in range(num_thresholds - 1):
            auroc[k] += 0.5 * (tpr_[j + 1] - tpr_[j]) * (tnr_[j + 1] + tnr_[j])
            auprc[k] += (tpr_[j + 1] - tpr_[j]) * ppv_[j + 1]

    fnr = 1 - sens  # false negative rate, miss rate
    fpr = 1 - spec  # false positive rate, fall-out
    fdr = 1 - prec  # false discovery rate
    for_ = 1 - npv  # false omission rate
    plr = sens / fpr  # positive likelihood ratio
    nlr = fnr / spec  # negative likelihood ratio
    pt = np.sqrt(fpr) / (np.sqrt(sens) + np.sqrt(fpr))  # prevalence threshold
    ba = (sens + spec) / 2  # balanced accuracy
    f1 = 2 * sens * prec / (sens + prec)  # f1-measure
    fm = np.sqrt(prec * sens)  # fowlkes-mallows index
    bm = sens + spec - 1  # informedness, bookmaker informedness
    mk = prec + npv - 1  # markedness
    dor = plr / nlr  # diagnostic odds ratio

    if weights is None:
        _weights = np.ones(num_classes)
    else:
        _weights = weights / np.mean(weights)
    metrics = {}
    for m in [
        "sens",  # sensitivity, recall, hit rate, or true positive rate
        "spec",  # specificity, selectivity or true negative rate
        "prec",  # precision or positive predictive value
        "npv",  # negative predictive value
        "jac",  # jaccard index, threat score, or critical success index
        "acc",  # accuracy
        "phi",  # phi coefficient, or matthews correlation coefficient
        "fnr",  # false negative rate, miss rate
        "fpr",  # false positive rate, fall-out
        "fdr",  # false discovery rate
        "for_",  # false omission rate
        "plr",  # positive likelihood ratio
        "nlr",  # negative likelihood ratio
        "pt",  # prevalence threshold
        "ba",  # balanced accuracy
        "f1",  # f1-measure
        "fm",  # fowlkes-mallows index
        "bm",  # bookmaker informedness
        "mk",  # markedness
        "dor",  # diagnostic odds ratio
        "auroc",  # area under the receiver-operater characteristic curve (ROC AUC)
        "auprc",  # area under the precision-recall curve
    ]:
        metrics[m.strip("_")] = eval(m)
        metrics[f"macro_{m}".strip("_")] = (
            np.nanmean(eval(m) * _weights)
            if np.any(np.isfinite(eval(m)))
            else float("nan")
        )
    return metrics


@add_docstring(
    _METRICS_FROM_CONFUSION_MATRIX_PARAMS.format(
        metric="F1-measure", metrics="F1-measures"
    ),
    "prepend",
)
def f_measure(
    labels: Union[np.ndarray, Tensor],
    outputs: Union[np.ndarray, Tensor],
    num_classes: Optional[int] = None,
    weights: Optional[Union[np.ndarray, Tensor]] = None,
    thr: float = 0.5,
) -> Tuple[float, np.ndarray]:
    """
    Returns
    -------
    macro_f1 : float
        Macro-averaged F1-measure.
    f1 : numpy.ndarray
        F1-measures for each class, of shape: ``(n_classes,)``.

    """
    m = metrics_from_confusion_matrix(labels, outputs, num_classes, weights, thr)

    return m["macro_f1"], m["f1"]


@add_docstring(
    _METRICS_FROM_CONFUSION_MATRIX_PARAMS.format(
        metric="sensitivity", metrics="sensitivities"
    ),
    "prepend",
)
def sensitivity(
    labels: Union[np.ndarray, Tensor],
    outputs: Union[np.ndarray, Tensor],
    num_classes: Optional[int] = None,
    weights: Optional[Union[np.ndarray, Tensor]] = None,
    thr: float = 0.5,
) -> Tuple[float, np.ndarray]:
    """
    Returns
    -------
    macro_sens : float
        Macro-averaged sensitivity.
    sens : numpy.ndarray
        Sensitivities for each class, of shape ``(n_classes,)``.

    """
    m = metrics_from_confusion_matrix(labels, outputs, num_classes, weights, thr)

    return m["macro_sens"], m["sens"]


# aliases
recall = sensitivity
true_positive_rate = sensitivity
hit_rate = sensitivity


@add_docstring(
    _METRICS_FROM_CONFUSION_MATRIX_PARAMS.format(
        metric="precision", metrics="precisions"
    ),
    "prepend",
)
def precision(
    labels: Union[np.ndarray, Tensor],
    outputs: Union[np.ndarray, Tensor],
    num_classes: Optional[int] = None,
    weights: Optional[Union[np.ndarray, Tensor]] = None,
    thr: float = 0.5,
) -> Tuple[float, np.ndarray]:
    """
    Returns
    -------
    macro_prec : float
        Macro-averaged precision.
    prec : numpy.ndarray
        Precisions for each class, of shape ``(n_classes,)``.

    """
    m = metrics_from_confusion_matrix(labels, outputs, num_classes, weights, thr)

    return m["macro_prec"], m["prec"]


# aliases
positive_predictive_value = precision


@add_docstring(
    _METRICS_FROM_CONFUSION_MATRIX_PARAMS.format(
        metric="specificity", metrics="specificities"
    ),
    "prepend",
)
def specificity(
    labels: Union[np.ndarray, Tensor],
    outputs: Union[np.ndarray, Tensor],
    num_classes: Optional[int] = None,
    weights: Optional[Union[np.ndarray, Tensor]] = None,
    thr: float = 0.5,
) -> Tuple[float, np.ndarray]:
    """
    Returns
    -------
    macro_spec : float
        Macro-averaged specificity.
    spec : numpy.ndarray
        Specificities for each class, of shape ``(n_classes,)``.

    """
    m = metrics_from_confusion_matrix(labels, outputs, num_classes, weights, thr)

    return m["macro_spec"], m["spec"]


# aliases
selectivity = specificity
true_negative_rate = specificity


@add_docstring(
    _METRICS_FROM_CONFUSION_MATRIX_PARAMS.format(
        metric="AUROC and macro AUPRC", metrics="AUPRCs, AUPRCs"
    ),
    "prepend",
)
def auc(
    labels: Union[np.ndarray, Tensor],
    outputs: Union[np.ndarray, Tensor],
    num_classes: Optional[int] = None,
    weights: Optional[Union[np.ndarray, Tensor]] = None,
    thr: float = 0.5,
) -> Tuple[float, float, np.ndarray, np.ndarray]:
    """
    Returns
    -------
    macro_auroc : float
        Macro-averaged AUROC.
    macro_auprc : float
        Macro-averaged AUPRC.
    auprc : numpy.ndarray
        AUPRCs for each class, of shape ``(n_classes,)``.
    auprc : numpy.ndarray
        AUPRCs for each class, of shape ``(n_classes,)``.

    """
    if outputs.ndim == 1:
        raise ValueError(
            "outputs must be of shape (n_samples, n_classes) to compute AUC"
        )
    m = metrics_from_confusion_matrix(labels, outputs, num_classes, weights, thr)

    return m["macro_auroc"], m["macro_auprc"], m["auroc"], m["auprc"]


@add_docstring(
    _METRICS_FROM_CONFUSION_MATRIX_PARAMS.format(
        metric="accuracy", metrics="accuracies"
    ),
    "prepend",
)
def accuracy(
    labels: Union[np.ndarray, Tensor],
    outputs: Union[np.ndarray, Tensor],
    num_classes: Optional[int] = None,
    weights: Optional[Union[np.ndarray, Tensor]] = None,
    thr: float = 0.5,
) -> float:
    """
    Returns
    -------
    macro_acc : float
        Macro-averaged accuracy.
    acc: numpy.ndarray,
        Accuracies for each class, of shape ``(n_classes,)``.

    """
    m = metrics_from_confusion_matrix(labels, outputs, num_classes, weights)

    return m["macro_acc"], m["acc"]


def QRS_score(
    rpeaks_truths: Sequence[Union[np.ndarray, Sequence[int]]],
    rpeaks_preds: Sequence[Union[np.ndarray, Sequence[int]]],
    fs: Real,
    thr: float = 0.075,
) -> float:
    """
    QRS accuracy score, proposed in CPSC2019.

    Parameters
    ----------
    rpeaks_truths : array_like
        array of ground truths of rpeaks locations (indices)
        from multiple records.
    rpeaks_preds : array_like
        predictions of ground truths of rpeaks locations (indices)
        for multiple records.
    fs : numbers.Real
        Sampling frequency of ECG signal
    thr : float, default 0.075
        Threshold for a prediction to be truth positive,
        with units in seconds.

    Returns
    -------
    rec_acc : float
        Accuracy of predictions.

    """
    assert len(rpeaks_truths) == len(
        rpeaks_preds
    ), f"number of records does not match, truth indicates {len(rpeaks_truths)}, while pred indicates {len(rpeaks_preds)}"
    n_records = len(rpeaks_truths)
    record_flags = np.ones((len(rpeaks_truths),), dtype=float)
    thr_ = thr * fs

    for idx, (truth_arr, pred_arr) in enumerate(zip(rpeaks_truths, rpeaks_preds)):
        false_negative = 0
        false_positive = 0
        true_positive = 0
        extended_truth_arr = np.concatenate((truth_arr.astype(int), [int(9.5 * fs)]))
        for j, t_ind in enumerate(extended_truth_arr[:-1]):
            next_t_ind = extended_truth_arr[j + 1]
            loc = np.where(np.abs(pred_arr - t_ind) <= thr_)[0]
            if j == 0:
                err = np.where(
                    (pred_arr >= 0.5 * fs + thr_) & (pred_arr <= t_ind - thr_)
                )[0]
            else:
                err = np.array([], dtype=int)
            err = np.append(
                err,
                np.where((pred_arr >= t_ind + thr_) & (pred_arr <= next_t_ind - thr_))[
                    0
                ],
            )

            false_positive += len(err)
            if len(loc) >= 1:
                true_positive += 1
                false_positive += len(loc) - 1
            elif len(loc) == 0:
                false_negative += 1

        if false_negative + false_positive > 1:
            record_flags[idx] = 0
        elif false_negative == 1 and false_positive == 0:
            record_flags[idx] = 0.3
        elif false_negative == 0 and false_positive == 1:
            record_flags[idx] = 0.7

    rec_acc = round(np.sum(record_flags) / n_records, 4)

    return rec_acc


def cls_to_bin(
    labels: Union[np.ndarray, Tensor],
    outputs: Union[np.ndarray, Tensor],
    num_classes: Optional[int] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Convert categorical (of shape ``(n_samples,)``) labels and outputs
    to binary (of shape ``(n_samples, n_classes)``) labels and outputs if applicable.

    Parameters
    ----------
    labels : numpy.ndarray or torch.Tensor
        Categorical labels of shape ``(n_samples,)``,
        or binary labels of shape ``(n_samples, n_classes)``.
    outputs : numpy.ndarray or torch.Tensor
        Categorical outputs of shape ``(n_samples,)``,
        or binary outputs of shape ``(n_samples, n_classes)``.
    num_classes : int, optional
        Number of classes.
        Required if both `labels` and `outputs` are categorical.

    Returns
    -------
    labels : numpy.ndarray
        Binary labels of shape ``(n_samples, n_classes)``.
    outputs: numpy.ndarray
        Binary outputs of shape ``(n_samples, n_classes)``.

    """
    if isinstance(labels, Tensor):
        labels = labels.cpu().numpy()
    if isinstance(outputs, Tensor):
        outputs = outputs.cpu().numpy()
    if labels.ndim == outputs.ndim == 1:
        assert (
            num_classes is not None
        ), "num_classes is required if both labels and outputs are categorical"
        shape = (labels.shape[0], num_classes)
        labels = _cls_to_bin(labels, shape)
        outputs = _cls_to_bin(outputs, shape)
    elif labels.ndim == 1:
        shape = outputs.shape
        labels = _cls_to_bin(labels, shape)
    elif outputs.ndim == 1:
        shape = labels.shape
        outputs = _cls_to_bin(outputs, shape)
    return labels, outputs


def _cls_to_bin(cls: np.ndarray, shape: Tuple[int]) -> np.ndarray:
    """Convert categorical array to binary array.

    Parameters
    ----------
    cls : numpy.ndarray
        Categorical array of shape ``(n_samples,)``.
    shape: tuple,
        Shape of binary array.

    Returns
    -------
    numpy.ndarray
        Binarized array of `cls` of shape ``(n_samples, n_classes)``.

    """
    bin_ = np.zeros(shape)
    for i in range(shape[0]):
        bin_[i, cls[i]] = 1
    return bin_


def compute_wave_delineation_metrics(
    truth_masks: Sequence[np.ndarray],
    pred_masks: Sequence[np.ndarray],
    class_map: Dict[str, int],
    fs: Real,
    mask_format: str = "channel_first",
    tol: Real = 0.15,
) -> Dict[str, Dict[str, float]]:
    f"""
    Compute metrics for the task of ECG wave delineation
    (sensitivity, precision, f1_score, mean error and standard deviation of the mean errors)
    for multiple evaluations.

    Parameters
    ----------
    truth_masks : Sequence[numpy.ndarray]
        A sequence of ground truth masks,
        each of which can also hold multiple masks from different samples
        (differ by record or by lead).
        Each mask is of shape ``(n_channels, n_timesteps)``
        or ``(n_timesteps, n_channels)``.
    pred_masks : Sequence[numpy.ndarray]
        Predictions corresponding to `truth_masks`,
        of the same shapes.
    class_map : dict
        Class map, mapping names to waves to numbers from 0 to n_classes-1,
        the keys should contain {", ".join([f'"{item}"' for item in ECGWaveFormNames])}.
    fs : numbers.Real
        Sampling frequency of the signal corresponding to the masks,
        used to compute the duration of each waveform,
        and thus the error and standard deviations of errors.
    mask_format : str, default "channel_first"
        Format of the mask, one of the following:
        "channel_last" (alias "lead_last"), or
        "channel_first" (alias "lead_first")
    tol : float, default 0.15
        Tolerance for the duration of the waveform,
        with units in seconds.

    Returns
    -------
    scorings : dict
        scorings of onsets and offsets of pwaves, qrs complexes, twaves.
        Each scoring is a dict consisting of the following metrics:
        sensitivity, precision, f1_score, mean_error, standard_deviation

    """
    assert len(truth_masks) == len(
        pred_masks
    ), "length of truth_masks and pred_masks should be the same"
    truth_waveforms, pred_waveforms = [], []
    # compute for each element
    for tm, pm in zip(truth_masks, pred_masks):
        n_masks = (
            tm.shape[0]
            if mask_format.lower() in ["channel_first", "lead_first"]
            else tm.shape[1]
        )

        new_t = masks_to_waveforms(tm, class_map, fs, mask_format)
        new_t = [
            new_t[f"lead_{idx+1}"] for idx in range(n_masks)
        ]  # list of list of `ECGWaveForm`s
        truth_waveforms += new_t

        new_p = masks_to_waveforms(pm, class_map, fs, mask_format)
        new_p = [
            new_p[f"lead_{idx+1}"] for idx in range(n_masks)
        ]  # list of list of `ECGWaveForm`s
        pred_waveforms += new_p

    scorings = compute_metrics_waveform(truth_waveforms, pred_waveforms, fs, tol)

    return scorings


def compute_metrics_waveform(
    truth_waveforms: Sequence[Sequence[ECGWaveForm]],
    pred_waveforms: Sequence[Sequence[ECGWaveForm]],
    fs: Real,
    tol: Real = 0.15,
) -> Dict[str, Dict[str, float]]:
    """
    compute the sensitivity, precision, f1_score, mean error
    and standard deviation of the mean errors,
    of evaluations on a multiple samples (differ by records, or leads).

    Parameters
    ----------
    truth_waveforms : Sequence[Sequence[ECGWaveForm]]
        The ground truth,
        each element is a sequence of `ECGWaveForm` from the same sample.
    pred_waveforms : Sequence[Sequence[ECGWaveForm]]
        The predictions corresponding to `truth_waveforms`,
        each element is a sequence of :class:`ECGWaveForm` from the same sample.
    fs : numbers.Real
        Sampling frequency of the signal corresponding to the waveforms,
        used to compute the duration of each waveform,
        and thus the error and standard deviations of errors.
    tol : float, default 0.15
        Tolerance for the duration of the waveform,
        with units in seconds.

    Returns
    -------
    scorings : dict
        Scorings of onsets and offsets of pwaves, qrs complexes, twaves.
        Each scoring is a dict consisting of the following metrics:
        sensitivity, precision, f1_score, mean_error, standard_deviation.

    """
    truth_positive = dict(
        {
            f"{wave}_{term}": 0
            for wave in ECGWaveFormNames
            for term in ["onset", "offset"]
        }
    )
    false_positive = dict(
        {
            f"{wave}_{term}": 0
            for wave in ECGWaveFormNames
            for term in ["onset", "offset"]
        }
    )
    false_negative = dict(
        {
            f"{wave}_{term}": 0
            for wave in ECGWaveFormNames
            for term in ["onset", "offset"]
        }
    )
    errors = dict(
        {
            f"{wave}_{term}": []
            for wave in ECGWaveFormNames
            for term in ["onset", "offset"]
        }
    )
    # accumulating results
    for tw, pw in zip(truth_waveforms, pred_waveforms):
        s = _compute_metrics_waveform(tw, pw, fs, tol)
        for wave in [
            "pwave",
            "qrs",
            "twave",
        ]:
            for term in ["onset", "offset"]:
                truth_positive[f"{wave}_{term}"] += s[f"{wave}_{term}"][
                    "truth_positive"
                ]
                false_positive[f"{wave}_{term}"] += s[f"{wave}_{term}"][
                    "false_positive"
                ]
                false_negative[f"{wave}_{term}"] += s[f"{wave}_{term}"][
                    "false_negative"
                ]
                errors[f"{wave}_{term}"] += s[f"{wave}_{term}"]["errors"]
    scorings = dict()
    for wave in ECGWaveFormNames:
        for term in ["onset", "offset"]:
            tp = truth_positive[f"{wave}_{term}"]
            fp = false_positive[f"{wave}_{term}"]
            fn = false_negative[f"{wave}_{term}"]
            err = errors[f"{wave}_{term}"]
            sensitivity = tp / (tp + fn + DEFAULTS.eps)
            precision = tp / (tp + fp + DEFAULTS.eps)
            f1_score = (
                2 * sensitivity * precision / (sensitivity + precision + DEFAULTS.eps)
            )
            mean_error = np.mean(err) * 1000 / fs if len(err) > 0 else np.nan
            standard_deviation = np.std(err) * 1000 / fs if len(err) > 0 else np.nan
            scorings[f"{wave}_{term}"] = dict(
                sensitivity=sensitivity,
                precision=precision,
                f1_score=f1_score,
                mean_error=mean_error,
                standard_deviation=standard_deviation,
            )

    return scorings


def _compute_metrics_waveform(
    truths: Sequence[ECGWaveForm],
    preds: Sequence[ECGWaveForm],
    fs: Real,
    tol: Real = 0.15,
) -> Dict[str, Dict[str, float]]:
    """
    Compute the sensitivity, precision, f1_score, mean error
    and standard deviation of the mean errors,
    of evaluations on a single sample (the same record, the same lead)

    Parameters
    ----------
    truths : Sequence[ECGWaveForm]
        The ground truth
    preds : Sequence[ECGWaveForm]
        The predictions corresponding to `truths`.
    fs : numbers.Real
        Sampling frequency of the signal corresponding to the waveforms,
        used to compute the duration of each waveform,
        and thus the error and standard deviations of errors.
    tol : float, default 0.15,
        Tolerance for the duration of the waveform,
        with units in seconds.

    Returns
    -------
    scorings : dict
        Scorings of onsets and offsets of pwaves, qrs complexes, twaves.
        Each scoring is a dict consisting of the following metrics:
        truth_positive, false_negative, false_positive, errors,
        sensitivity, precision, f1_score, mean_error, standard_deviation

    """
    pwave_onset_truths, pwave_offset_truths, pwave_onset_preds, pwave_offset_preds = (
        [],
        [],
        [],
        [],
    )
    qrs_onset_truths, qrs_offset_truths, qrs_onset_preds, qrs_offset_preds = (
        [],
        [],
        [],
        [],
    )
    twave_onset_truths, twave_offset_truths, twave_onset_preds, twave_offset_preds = (
        [],
        [],
        [],
        [],
    )

    for item in ["truths", "preds"]:
        for w in eval(item):
            for term in ["onset", "offset"]:
                eval(f"{w.name}_{term}_{item}.append(w.{term})")

    scorings = dict()
    for wave in ECGWaveFormNames:
        for term in ["onset", "offset"]:
            (
                truth_positive,
                false_negative,
                false_positive,
                errors,
                sensitivity,
                precision,
                f1_score,
                mean_error,
                standard_deviation,
            ) = _compute_metrics_base(
                eval(f"{wave}_{term}_truths"), eval(f"{wave}_{term}_preds"), fs, tol
            )
            scorings[f"{wave}_{term}"] = dict(
                truth_positive=truth_positive,
                false_negative=false_negative,
                false_positive=false_positive,
                errors=errors,
                sensitivity=sensitivity,
                precision=precision,
                f1_score=f1_score,
                mean_error=mean_error,
                standard_deviation=standard_deviation,
            )
    return scorings


def _compute_metrics_base(
    truths: Sequence[Real], preds: Sequence[Real], fs: Real, tol: Real = 0.15
) -> Dict[str, float]:
    r"""Base function for computing the metrics of the onset and offset of a waveform.

    Parameters
    ----------
    truths : Sequence[numbers.Real]
        Ground truth of indices of corresponding critical points.
    preds : Sequence[numbers.Real]
        Predicted indices of corresponding critical points.
    fs : numbers.Real
        Sampling frequency of the signal corresponding to the critical points,
        used to compute the duration of each waveform,
        and thus the error and standard deviations of errors.
    tol : float, default 0.15
        Tolerance for the duration of the waveform,
        with units in seconds.

    Returns
    -------
    tuple
        Tuple of metrics:
        truth_positive, false_negative, false_positive, errors,
        sensitivity, precision, f1_score, mean_error, standard_deviation.
        See [#dl_seg]_ for more details.

    References
    ----------
    .. [#dl_seg] Moskalenko, Viktor, Nikolai Zolotykh, and Grigory Osipov.
                 "Deep Learning for ECG Segmentation." International Conference on Neuroinformatics.
                 Springer, Cham, 2019.

    """
    _tolerance = round(tol * fs)
    _truths = np.array(truths)
    _preds = np.array(preds)
    truth_positive, false_positive, false_negative = 0, 0, 0
    errors = []
    n_included = 0
    for point in truths:
        _pred = _preds[np.where(np.abs(_preds - point) <= _tolerance)[0].tolist()]
        if len(_pred) > 0:
            truth_positive += 1
            idx = np.argmin(np.abs(_pred - point))
            errors.append(_pred[idx] - point)
        else:
            false_negative += 1
        n_included += len(_pred)

    false_positive = len(_preds) - truth_positive

    sensitivity = truth_positive / (truth_positive + false_negative + DEFAULTS.eps)
    precision = truth_positive / (truth_positive + false_positive + DEFAULTS.eps)
    f1_score = 2 * sensitivity * precision / (sensitivity + precision + DEFAULTS.eps)
    mean_error = np.mean(errors) * 1000 / fs if len(errors) > 0 else np.nan
    standard_deviation = np.std(errors) * 1000 / fs if len(errors) > 0 else np.nan

    return (
        truth_positive,
        false_negative,
        false_positive,
        errors,
        sensitivity,
        precision,
        f1_score,
        mean_error,
        standard_deviation,
    )



"""
torch_ecg.models.loss
=====================
Custom loss functions for ECG analysis, as a complement to
built-in loss functions in PyTorch.

.. contents:: torch_ecg.models
    :depth: 1
    :local:
    :backlinks: top

.. currentmodule:: torch_ecg.models.loss

.. autosummary::
    :toctree: generated/
    :recursive:

    WeightedBCELoss
    BCEWithLogitsWithClassWeightLoss
    MaskedBCEWithLogitsLoss
    FocalLoss
    AsymmetricLoss

"""

from numbers import Real
from typing import Any, Optional

import torch
import torch.nn.functional as F
from torch import Tensor, nn


__all__ = [
    "WeightedBCELoss",
    "BCEWithLogitsWithClassWeightLoss",
    "MaskedBCEWithLogitsLoss",
    "FocalLoss",
    "AsymmetricLoss",
]


def weighted_binary_cross_entropy(
    sigmoid_x: Tensor,
    targets: Tensor,
    pos_weight: Tensor,
    weight: Optional[Tensor] = None,
    size_average: bool = True,
    reduce: bool = True,
) -> Tensor:
    """Weighted Binary Cross Entropy Loss function.

    This implementation is based on [#wbce]_.

    Parameters
    ----------
    sigmoid_x : torch.Tensor
        Predicted probability of size ``[N, C]``, N sample and C Class.
        Eg. Must be in range of ``[0, 1]``,
        i.e. output from :class:`~torch.nn.Sigmoid`.
    targets : torch.Tensor
        True value, one-hot-like vector of size ``[N, C]``.
    pos_weight : torch.Tensor
        Weight for postive sample.
    weight : torch.Tensor, optional
        Weight for each class, of size ``[1, C]``.
    size_average : bool, default True
        If True, the losses are averaged
        over each loss element in the batch.
        Valid only if `reduce` is True.
    reduce : bool, default True
        If True, the losses are averaged or summed
        over observations for each minibatch.

    Returns
    -------
    loss : torch.Tensor
        Weighted Binary Cross Entropy Loss.

    References
    ----------
    .. [#wbce] https://github.com/pytorch/pytorch/issues/5660#issuecomment-403770305

    """
    if not (targets.size() == sigmoid_x.size()):
        raise ValueError(
            f"Target size ({targets.size()}) must be the same as input size ({sigmoid_x.size()})"
        )

    loss = (
        -pos_weight * targets * sigmoid_x.log() - (1 - targets) * (1 - sigmoid_x).log()
    )
    # print(pos_weight, targets, sigmoid_x)

    if weight is not None:
        loss = loss * weight

    if not reduce:
        return loss
    elif size_average:
        return loss.mean()
    else:
        return loss.sum()


class WeightedBCELoss(nn.Module):
    """Weighted Binary Cross Entropy Loss class.

    This implementation is based on [#wbce]_.

    Parameters
    ----------
    pos_weight : torch.Tensor
        Weight for postive sample.
    weight : torch.Tensor, optional
        Weight for each class, of size ``[1, C]``.
    PosWeightIsDynamic : bool, default False
        If True, the pos_weight is computed on each batch.
        If `pos_weight` is None, then it remains None.
    WeightIsDynamic : bool, default False
        If True, the weight is computed on each batch.
        If `weight` is None, then it remains None.
    size_average : bool, default True
        If True, the losses are averaged
        over each loss element in the batch.
        Valid only if `reduce` is True.
    reduce : bool, default True
        If True, the losses are averaged or summed
        over observations for each minibatch.

    References
    ----------
    .. [#wbce] https://github.com/pytorch/pytorch/issues/5660#issuecomment-403770305

    """

    __name__ = "WeightedBCELoss"

    def __init__(
        self,
        pos_weight: Tensor,
        weight: Optional[Tensor] = None,
        PosWeightIsDynamic: bool = False,
        WeightIsDynamic: bool = False,
        size_average: bool = True,
        reduce: bool = True,
    ) -> None:
        super().__init__()

        self.register_buffer("pos_weight", pos_weight)
        if weight is None:
            weight = torch.ones_like(pos_weight)
        self.register_buffer("weight", weight)
        self.size_average = size_average
        self.reduce = reduce
        self.PosWeightIsDynamic = PosWeightIsDynamic

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        input : torch.Tensor
            The predicted probability tensor,
            of shape ``(batch_size, ..., n_classes)``.
        target : torch.Tensor
            The target tensor,
            of shape ``(batch_size, ..., n_classes)``.

        Returns
        -------
        loss : torch.Tensor
            The weighted binary cross entropy loss.

        """
        if self.PosWeightIsDynamic:
            positive_counts = target.sum(dim=0, keepdim=True)
            nBatch = len(target)
            self.pos_weight = (nBatch - positive_counts) / (positive_counts + 1e-7)

        return weighted_binary_cross_entropy(
            input,
            target,
            pos_weight=self.pos_weight,
            weight=self.weight,
            size_average=self.size_average,
            reduce=self.reduce,
        )


class BCEWithLogitsWithClassWeightLoss(nn.BCEWithLogitsLoss):
    """Class-weighted Binary Cross Entropy Loss class.

    Parameters
    ----------
    class_weight : torch.Tensor
        Class weight, of shape ``(1, n_classes)``.

    """

    __name__ = "BCEWithLogitsWithClassWeightLoss"

    def __init__(self, class_weight: Tensor) -> None:
        super().__init__(reduction="none")
        self.register_buffer("class_weight", class_weight)

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        input : torch.Tensor
            The predicted value tensor (before sigmoid),
            of shape ``(batch_size, ..., n_classes)``.
        target : torch.Tensor
            The target tensor,
            of shape ``(batch_size, ..., n_classes)``.

        Returns
        -------
        torch.Tensor
            The class-weighted binary cross entropy loss.

        """
        loss = super().forward(input, target)
        loss = torch.mean(loss * self.class_weight)
        return loss


class MaskedBCEWithLogitsLoss(nn.BCEWithLogitsLoss):
    """Masked Binary Cross Entropy Loss class.

    This loss is used mainly for the segmentation task, where
    there are some regions that are of much higher importance,
    for example, the onsets and offsets of some particular events
    (e.g. paroxysmal atrial fibrillation (AF) episodes).

    This loss is proposed in [#mbce]_, with a reference to the loss
    function used in the U-Net paper [#unet]_.


    References
    ----------
    .. [#mbce] Wen, Hao, and Jingsu Kang. "A comparative study on neural networks for
               paroxysmal atrial fibrillation events detection from electrocardiography."
               Journal of Electrocardiology 75 (2022): 19-27.
    .. [#unet] Ronneberger, Olaf, Philipp Fischer, and Thomas Brox. "U-net: Convolutional
               networks for biomedical image segmentation." International Conference on
               Medical image computing and computer-assisted intervention. Springer, Cham,
               2015.

    """

    __name__ = "MaskedBCEWithLogitsLoss"

    def __init__(self) -> None:
        super().__init__(reduction="none")

    def forward(self, input: Tensor, target: Tensor, weight_mask: Tensor) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        input : torch.Tensor
            The predicted value tensor (before sigmoid),
            of shape ``(batch_size, sig_len, n_classes)``.
        target : torch.Tensor
            The target tensor,
            of shape ``(batch_size, sig_len, n_classes)``.
        weight_mask: torch.Tensor
            The weight mask tensor,
            of shape ``(batch_size, sig_len, n_classes)``.

        Returns
        -------
        torch.Tensor
            The masked binary cross entropy loss.

        NOTE
        ----
        `input`, `target`, and `weight_mask` should be
        3-D tensors of the same shape.

        """
        loss = super().forward(input, target)
        loss = torch.mean(loss * weight_mask)
        return loss


class FocalLoss(nn.modules.loss._WeightedLoss):
    """Focal loss class.

    The focal loss is proposed in [1]_, and this implementation is
    based on [2]_, [3]_, and [4]_. The focal loss is computed as follows:

    .. math::

        \\operatorname{FL}(p_t) = -\\alpha_t (1 - p_t)^{\\gamma} \\, \\log(p_t)

    Where:

       - :math:`p_t` is the model's estimated probability for each class.

    Parameters
    ----------
    gamma : float, default 2.0
        The gamma parameter of focal loss.
    weight : torch.Tensor, optional
        If `multi_label` is True,
        is a manual rescaling weight given to the loss of each batch element,
        of size ``batch_size``;
        if `multi_label` is False,
        is a weight for each class, of size ``n_classes``.
    class_weight : torch.Tensor, optional
        The class weight, of shape ``(1, n_classes)``.
    size_average : bool, optional
        Not used, to keep in accordance with PyTorch native loss.
    reduce : bool, optional
        Not used, to keep in accordance with PyTorch native loss.
    reduction: {"none", "mean", "sum"}, optional
        Specifies the reduction to apply to the output, by default "mean".
    multi_label : bool, default True
        If True, the loss is computed for multi-label classification.

    References
    ----------
    .. [1] Lin, Tsung-Yi, et al. "Focal loss for dense object detection."
           Proceedings of the IEEE international conference on computer vision. 2017.
    .. [2] https://github.com/kornia/kornia/blob/master/kornia/losses/focal.py
    .. [3] https://github.com/clcarwin/focal_loss_pytorch/blob/master/focalloss.py
    .. [4] https://discuss.pytorch.org/t/is-this-a-correct-implementation-for-focal-loss-in-pytorch/43327

    """

    __name__ = "FocalLoss"

    def __init__(
        self,
        gamma: float = 2.0,
        weight: Optional[Tensor] = None,
        class_weight: Optional[Tensor] = None,  # alpha
        size_average: Optional[bool] = None,
        reduce: Optional[bool] = None,
        reduction: str = "mean",
        multi_label: bool = True,
        **kwargs: Any,
    ) -> None:
        if multi_label or weight is not None:
            w = weight
        else:
            w = class_weight
        if not multi_label and w.ndim == 2:
            w = w.squeeze(0)
        super().__init__(
            weight=w, size_average=size_average, reduce=reduce, reduction=reduction
        )
        # In practice `alpha` may be set by inverse class frequency or treated as a hyperparameter
        # the `class_weight` are usually inverse class frequencies
        # self.alpha = alpha
        self.gamma = gamma
        if multi_label:
            self.entropy_func = F.binary_cross_entropy_with_logits
            # for `binary_cross_entropy_with_logits`,
            # its parameter `weight` is a manual rescaling weight given to the loss of each batch element
            self.register_buffer("class_weight", class_weight)
        else:
            self.entropy_func = F.cross_entropy
            # for `cross_entropy`,
            # its parameter `weight` is a manual rescaling weight given to each class
            self.class_weight = None

    @property
    def alpha(self) -> Tensor:
        return self.class_weight

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        input : torch.Tensor
            The predicted value tensor (before sigmoid),
            of shape ``(batch_size, n_classes)``.
        target : torch.Tensor
            Multi-label binarized vector of shape ``(batch_size, n_classes)``,
            or single label binarized vector of shape ``(batch_size,)``.

        Returns
        -------
        torch.Tensor
            The focal loss.

        """
        entropy = self.entropy_func(
            input,
            target,
            weight=self.weight,
            reduction="none",
        )
        p_t = torch.exp(-entropy)
        fl = torch.pow(1 - p_t, self.gamma) * entropy
        if self.class_weight is not None:
            fl = fl * self.class_weight
        if self.reduction == "mean":
            fl = fl.mean()
        elif self.reduction == "sum":
            fl = fl.sum()
        return fl


class AsymmetricLoss(nn.Module):
    """Asymmetric loss class.

    The asymmetric loss is proposed in [#al]_, with official
    implementation in [#al_code]_. The asymmetric loss is defined as

    .. math::

        ASL = \\begin{cases}
            L_+ := (1-p)^{\\gamma_+} \\log(p) \\
            L_- := (p_m)^{\\gamma_-} \\log(1-p_m)
        \\end{cases}

    where :math:`p_m = \\max(p-m, 0)` is the shifted probability,
    with probability margin :math:`m`.
    The loss on one label of one sample is

    .. math::

        L = -yL_+ - (1-y)L_-

    Parameters
    ----------
    gamma_neg : numbers.Real, default 4
        Exponent of the multiplier to the negative loss.
    gamma_pos : numbers.Real, default 1
        Exponent of the multiplier to the positive loss.
    prob_margin : float, default 0.05
        The probability margin
    disable_torch_grad_focal_loss : bool, default False
        If True, disable :func:`torch.grad` for asymmetric focal loss computing.
    reduction : {"none", "mean", "sum"}, optional
        Specifies the reduction to apply to the output, by default "mean".
    implementation : {"alibaba-miil", "deep-psp"}, optional
        Implementation by Alibaba-MIIL, or by `DeepPSP`, case insensitive.

    NOTE
    ----
    Since :class:`AsymmetricLoss` aims at emphasizing the contribution of positive samples,
    `gamma_neg` is usually greater than `gamma_pos`.

    TODO
    ----
    1. Evaluate the settings that `gamma_neg`, `gamma_pos` are tensors,
       of shape ``(1, n_classes)``, in which case we would have one ratio
       of positive to negative for each class.

    References
    ----------
    .. [#al] Ridnik, Tal, et al. "Asymmetric Loss for Multi-Label Classification."
             Proceedings of the IEEE/CVF International Conference on Computer Vision. 2021.
    .. [#al_code] https://github.com/Alibaba-MIIL/ASL/

    """

    __name__ = "AsymmetricLoss"

    def __init__(
        self,
        gamma_neg: Real = 4,
        gamma_pos: Real = 1,
        prob_margin: float = 0.05,
        disable_torch_grad_focal_loss: bool = False,
        reduction: str = "mean",
        implementation: str = "alibaba-miil",
    ) -> None:
        super().__init__()
        self.implementation = implementation.lower()
        assert self.implementation in [
            "alibaba-miil",
            "deep-psp",
            "deeppsp",
        ]
        self.gamma_neg = gamma_neg
        self.gamma_pos = gamma_pos
        self.prob_margin = prob_margin
        if self.prob_margin < 0:
            raise ValueError("`prob_margin` must be non-negative")
        self.disable_torch_grad_focal_loss = disable_torch_grad_focal_loss
        self.eps = 1e-8
        self.reduction = reduction.lower()

        if self.implementation == "alibaba-miil":
            self.targets = (
                self.anti_targets
            ) = self.xs_pos = self.xs_neg = self.asymmetric_w = self.loss = None
        elif self.implementation in [
            "deep-psp",
            "deeppsp",
        ]:
            self.targets = (
                self.anti_targets
            ) = (
                self.xs_pos
            ) = self.xs_neg = self.loss = self.loss_pos = self.loss_neg = None

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        input : torch.Tensor
            The predicted value tensor,
            of shape ``(batch_size, n_classes)``.
        target : torch.Tensor
            The target tensor,
            of shape ``(batch_size, n_classes)``.

        Returns
        -------
        torch.Tensor
            The asymmetric loss.

        """
        if self.implementation == "alibaba-miil":
            return self._forward_alibaba_miil(input, target)
        else:
            return self._forward_deep_psp(input, target)

    def _forward_deep_psp(self, input: Tensor, target: Tensor) -> Tensor:
        """Forward pass of DeepPSP implementation.

        Parameters
        ----------
        input : torch.Tensor
            The predicted value tensor,
            of shape ``(batch_size, n_classes)``.
        target : torch.Tensor
            The target tensor,
            of shape ``(batch_size, n_classes)``.

        Returns
        -------
        torch.Tensor
            The asymmetric loss.

        """
        self.targets = target
        self.anti_targets = 1 - target

        # Calculating Probabilities
        self.xs_pos = torch.sigmoid(input)
        self.xs_neg = 1.0 - self.xs_pos

        # Asymmetric Clipping
        if self.prob_margin > 0:
            self.xs_neg.add_(self.prob_margin).clamp_(max=1)

        # Basic CE calculation
        self.loss_pos = self.targets * torch.log(self.xs_pos.clamp(min=self.eps))
        self.loss_neg = self.anti_targets * torch.log(self.xs_neg.clamp(min=self.eps))

        # Asymmetric Focusing
        if self.gamma_neg > 0 or self.gamma_pos > 0:
            if self.disable_torch_grad_focal_loss:
                prev = torch.is_grad_enabled()
                torch.set_grad_enabled(False)
            self.loss_pos *= torch.pow(1 - self.xs_pos, self.gamma_pos)
            self.loss_neg *= torch.pow(self.xs_pos, self.gamma_neg)
            if self.disable_torch_grad_focal_loss:
                torch.set_grad_enabled(prev)
        self.loss = -self.loss_pos - self.loss_neg

        if self.reduction == "mean":
            self.loss = self.loss.mean()
        elif self.reduction == "sum":
            self.loss = self.loss.sum()
        return self.loss

    def _forward_alibaba_miil(self, input: Tensor, target: Tensor) -> Tensor:
        """Forward pass of Alibaba MIIL implementation.

        Parameters
        ----------
        input : torch.Tensor
            The predicted value tensor,
            of shape ``(batch_size, n_classes)``.
        target : torch.Tensor
            The target tensor,
            of shape ``(batch_size, n_classes)``.

        Returns
        -------
        torch.Tensor
            The asymmetric loss.

        """
        self.targets = target
        self.anti_targets = 1 - target

        # Calculating Probabilities
        self.xs_pos = torch.sigmoid(input)
        self.xs_neg = 1.0 - self.xs_pos

        # Asymmetric Clipping
        if self.prob_margin > 0:
            self.xs_neg.add_(self.prob_margin).clamp_(max=1)

        # Basic CE calculation
        # loss = y * log(p) + (1-y) * log(1-p)
        self.loss = self.targets * torch.log(self.xs_pos.clamp(min=self.eps))
        self.loss.add_(self.anti_targets * torch.log(self.xs_neg.clamp(min=self.eps)))

        # Asymmetric Focusing
        if self.gamma_neg > 0 or self.gamma_pos > 0:
            if self.disable_torch_grad_focal_loss:
                prev = torch.is_grad_enabled()
                torch.set_grad_enabled(False)
            self.xs_pos = self.xs_pos * self.targets  # p * y
            self.xs_neg = self.xs_neg * self.anti_targets  # (1-p) * (1-y)
            self.asymmetric_w = torch.pow(
                1 - self.xs_pos - self.xs_neg,
                self.gamma_pos * self.targets + self.gamma_neg * self.anti_targets,
            )
            if self.disable_torch_grad_focal_loss:
                torch.set_grad_enabled(prev)
            self.loss *= self.asymmetric_w

        if self.reduction == "mean":
            self.loss = -self.loss.mean()
        elif self.reduction == "sum":
            self.loss = -self.loss.sum()
        else:
            self.loss = -self.loss
        return self.loss



# -*- coding: utf-8 -*-

from copy import deepcopy
from numbers import Real
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Union, Tuple

import numpy as np
import pandas as pd
import wfdb

from ...cfg import CFG, DEFAULTS
from ...utils import EAK
from ...utils.misc import add_docstring
from ...utils.utils_data import ECGWaveForm, masks_to_waveforms
from ..base import PhysioNetDataBase, DataBaseInfo


__all__ = [
    "LUDB",
    "compute_metrics",
]


_LUDB_INFO = DataBaseInfo(
    title="""
    Lobachevsky University Electrocardiography Database
    """,
    about="""
    1. consist of 200 10-second conventional 12-lead (i, ii, iii, avr, avl, avf, v1, v2, v3, v4, v5, v6) ECG signal records, with sampling frequency 500 Hz
    2. boundaries of P, T waves and QRS complexes were manually annotated by cardiologists, and with the corresponding diagnosis
    3. annotated are 16797 P waves, 21966 QRS complexes, 19666 T waves (in total, 58429 annotated waves)
    4. distributions of data:

        - rhythm distribution:

            +-------------------------+----------------+
            | Rhythms                 | Number of ECGs |
            +=========================+================+
            | Sinus rhythm            | 143            |
            +-------------------------+----------------+
            | Sinus tachycardia       | 4              |
            +-------------------------+----------------+
            | Sinus bradycardia       | 25             |
            +-------------------------+----------------+
            | Sinus arrhythmia        | 8              |
            +-------------------------+----------------+
            | Irregular sinus rhythm  | 2              |
            +-------------------------+----------------+
            | Abnormal rhythm         | 19             |
            +-------------------------+----------------+

        - electrical axis distribution:

            +----------------------------+----------------+
            | Heart electric axis        | Number of ECGs |
            +============================+================+
            | Normal                     | 75             |
            +----------------------------+----------------+
            | Left axis deviation (LAD)  | 66             |
            +----------------------------+----------------+
            | Vertical                   | 26             |
            +----------------------------+----------------+
            | Horizontal                 | 20             |
            +----------------------------+----------------+
            | Right axis deviation (RAD) | 3              |
            +----------------------------+----------------+
            | Undetermined               | 10             |
            +----------------------------+----------------+

        - distribution of records with conduction abnomalities (totally 79):

            +------------------------------------------------+----------------+
            | Conduction abnormalities                       | Number of ECGs |
            +================================================+================+
            | Sinoatrial blockade, undetermined              | 1              |
            +------------------------------------------------+----------------+
            | I degree AV block                              | 10             |
            +------------------------------------------------+----------------+
            | III degree AV-block                            | 5              |
            +------------------------------------------------+----------------+
            | Incomplete right bundle branch block           | 29             |
            +------------------------------------------------+----------------+
            | Incomplete left bundle branch block            | 6              |
            +------------------------------------------------+----------------+
            | Left anterior hemiblock                        | 16             |
            +------------------------------------------------+----------------+
            | Complete right bundle branch block             | 4              |
            +------------------------------------------------+----------------+
            | Complete left bundle branch block              | 4              |
            +------------------------------------------------+----------------+
            | Non-specific intravintricular conduction delay | 4              |
            +------------------------------------------------+----------------+

        - distribution of records with extrasystoles (totally 35):

            +------------------------------------------------------------------+----------------+
            | Extrasystoles                                                    | Number of ECGs |
            +==================================================================+================+
            | Atrial extrasystole, undetermined                                | 2              |
            +------------------------------------------------------------------+----------------+
            | Atrial extrasystole, low atrial                                  | 1              |
            +------------------------------------------------------------------+----------------+
            | Atrial extrasystole, left atrial                                 | 2              |
            +------------------------------------------------------------------+----------------+
            | Atrial extrasystole, SA-nodal extrasystole                       | 3              |
            +------------------------------------------------------------------+----------------+
            | Atrial extrasystole, type: single PAC                            | 4              |
            +------------------------------------------------------------------+----------------+
            | Atrial extrasystole, type: bigemini                              | 1              |
            +------------------------------------------------------------------+----------------+
            | Atrial extrasystole, type: quadrigemini                          | 1              |
            +------------------------------------------------------------------+----------------+
            | Atrial extrasystole, type: allorhythmic pattern                  | 1              |
            +------------------------------------------------------------------+----------------+
            | Ventricular extrasystole, morphology: polymorphic                | 2              |
            +------------------------------------------------------------------+----------------+
            | Ventricular extrasystole, localisation: RVOT, anterior wall      | 3              |
            +------------------------------------------------------------------+----------------+
            | Ventricular extrasystole, localisation: RVOT, antero-septal part | 1              |
            +------------------------------------------------------------------+----------------+
            | Ventricular extrasystole, localisation: IVS, middle part         | 1              |
            +------------------------------------------------------------------+----------------+
            | Ventricular extrasystole, localisation: LVOT, LVS                | 2              |
            +------------------------------------------------------------------+----------------+
            | Ventricular extrasystole, localisation: LV, undefined            | 1              |
            +------------------------------------------------------------------+----------------+
            | Ventricular extrasystole, type: single PVC                       | 6              |
            +------------------------------------------------------------------+----------------+
            | Ventricular extrasystole, type: intercalary PVC                  | 2              |
            +------------------------------------------------------------------+----------------+
            | Ventricular extrasystole, type: couplet                          | 2              |
            +------------------------------------------------------------------+----------------+

        - distribution of records with hypertrophies (totally 253):

            +-------------------------------+----------------+
            | Hypertrophies                 | Number of ECGs |
            +===============================+================+
            | Right atrial hypertrophy      | 1              |
            +-------------------------------+----------------+
            | Left atrial hypertrophy       | 102            |
            +-------------------------------+----------------+
            | Right atrial overload         | 17             |
            +-------------------------------+----------------+
            | Left atrial overload          | 11             |
            +-------------------------------+----------------+
            | Left ventricular hypertrophy  | 108            |
            +-------------------------------+----------------+
            | Right ventricular hypertrophy | 3              |
            +-------------------------------+----------------+
            | Left ventricular overload     | 11             |
            +-------------------------------+----------------+

        - distribution of records of pacing rhythms (totally 12):

            +------------------------------+----------------+
            | Cardiac pacing               | Number of ECGs |
            +==============================+================+
            | UNIpolar atrial pacing       | 1              |
            +------------------------------+----------------+
            | UNIpolar ventricular pacing  | 6              |
            +------------------------------+----------------+
            | BIpolar ventricular pacing   | 2              |
            +------------------------------+----------------+
            | Biventricular pacing         | 1              |
            +------------------------------+----------------+
            | P-synchrony                  | 2              |
            +------------------------------+----------------+

        - distribution of records with ischemia (totally 141):

            +-------------------------------------------------------+----------------+
            | Ischemia                                              | Number of ECGs |
            +=======================================================+================+
            | STEMI: anterior wall                                  | 8              |
            +-------------------------------------------------------+----------------+
            | STEMI: lateral wall                                   | 7              |
            +-------------------------------------------------------+----------------+
            | STEMI: septal                                         | 8              |
            +-------------------------------------------------------+----------------+
            | STEMI: inferior wall                                  | 1              |
            +-------------------------------------------------------+----------------+
            | STEMI: apical                                         | 5              |
            +-------------------------------------------------------+----------------+
            | Ischemia: anterior wall                               | 5              |
            +-------------------------------------------------------+----------------+
            | Ischemia: lateral wall                                | 8              |
            +-------------------------------------------------------+----------------+
            | Ischemia: septal                                      | 4              |
            +-------------------------------------------------------+----------------+
            | Ischemia: inferior wall                               | 10             |
            +-------------------------------------------------------+----------------+
            | Ischemia: posterior wall                              | 2              |
            +-------------------------------------------------------+----------------+
            | Ischemia: apical                                      | 6              |
            +-------------------------------------------------------+----------------+
            | Scar formation: lateral wall                          | 3              |
            +-------------------------------------------------------+----------------+
            | Scar formation: septal                                | 9              |
            +-------------------------------------------------------+----------------+
            | Scar formation: inferior wall                         | 3              |
            +-------------------------------------------------------+----------------+
            | Scar formation: posterior wall                        | 6              |
            +-------------------------------------------------------+----------------+
            | Scar formation: apical                                | 5              |
            +-------------------------------------------------------+----------------+
            | Undefined ischemia/scar/supp.NSTEMI: anterior wall    | 12             |
            +-------------------------------------------------------+----------------+
            | Undefined ischemia/scar/supp.NSTEMI: lateral wall     | 16             |
            +-------------------------------------------------------+----------------+
            | Undefined ischemia/scar/supp.NSTEMI: septal           | 5              |
            +-------------------------------------------------------+----------------+
            | Undefined ischemia/scar/supp.NSTEMI: inferior wall    | 3              |
            +-------------------------------------------------------+----------------+
            | Undefined ischemia/scar/supp.NSTEMI: posterior wall   | 4              |
            +-------------------------------------------------------+----------------+
            | Undefined ischemia/scar/supp.NSTEMI: apical           | 11             |
            +-------------------------------------------------------+----------------+

        - distribution of records with non-specific repolarization abnormalities (totally 85):

            +-------------------------------------------+----------------+
            | Non-specific repolarization abnormalities | Number of ECGs |
            +===========================================+================+
            | Anterior wall                             |      18        |
            +-------------------------------------------+----------------+
            | Lateral wall                              |      13        |
            +-------------------------------------------+----------------+
            | Septal                                    |      15        |
            +-------------------------------------------+----------------+
            | Inferior wall                             |      19        |
            +-------------------------------------------+----------------+
            | Posterior wall                            |      9         |
            +-------------------------------------------+----------------+
            | Apical                                    |      11        |
            +-------------------------------------------+----------------+

        - there are also 9 records with early repolarization syndrome

       there might well be records with multiple conditions.
    5. ludb.csv stores information about the subjects (gender, age, rhythm type, direction of the electrical axis of the heart, the presence of a cardiac pacemaker, etc.)
    6. Webpage of the database on PhysioNet [1]_. Paper describing the database [2]_.
    """,
    usage=[
        "ECG wave delineation",
        "ECG arrhythmia classification",
    ],
    issues="""
    1. (version 1.0.0, fixed in version 1.0.1) ADC gain might be wrong, either `units` should be μV, or `adc_gain` should be 1000 times larger
    """,
    references=[
        "https://physionet.org/content/ludb/1.0.1/",
        "Kalyakulina, A., Yusipov, I., Moskalenko, V., Nikolskiy, A., Kozlov, A., Kosonogov, K., Zolotykh, N., & Ivanchenko, M. (2020). Lobachevsky University Electrocardiography Database (version 1.0.1).",
    ],
    doi=[
        "10.1109/ACCESS.2020.3029211",
        "10.13026/eegm-h675",
    ],
)


@add_docstring(_LUDB_INFO.format_database_docstring(), mode="prepend")
class LUDB(PhysioNetDataBase):
    """
    Parameters
    ----------
    db_dir : str or pathlib.Path, optional
        Storage path of the database.
        If not specified, data will be fetched from Physionet.
    working_dir : str, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Level of logging verbosity.
    kwargs : dict, optional
        Auxilliary key word arguments.

    """

    __name__ = "LUDB"

    def __init__(
        self,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[Union[str, Path]] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            db_name="ludb",
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )
        if self.version == "1.0.0":
            self.logger.info(
                "Version of LUDB 1.0.0 has bugs, make sure that version 1.0.1 or higher is used"
            )
        self.fs = 500
        self.spacing = 1000 / self.fs
        self.data_ext = "dat"
        self.all_leads = deepcopy(EAK.Standard12Leads)
        self.all_leads_lower = [ld.lower() for ld in self.all_leads]
        # Version 1.0.0 has different beat_ann_ext: [f"atr_{item}" for item in self.all_leads_lower]
        self.beat_ann_ext = [f"{item}" for item in self.all_leads_lower]

        self._all_symbols = ["(", ")", "p", "N", "t"]
        """
        this can be obtained using the following code:
        >>> data_gen = LUDB(db_dir="/home/wenhao71/data/PhysioNet/ludb/1.0.1/")
        >>> all_symbols = set()
        >>> for rec in data_gen.all_records:
        ...     for ext in data_gen.beat_ann_ext:
        ...         ann = wfdb.rdann(data_gen.get_absolute_path(rec), extension=ext)
        ...         all_symbols.update(ann.symbol)
        """
        self._symbol_to_wavename = CFG(N="qrs", p="pwave", t="twave")
        self._wavename_to_symbol = CFG(
            {v: k for k, v in self._symbol_to_wavename.items()}
        )
        self.class_map = CFG(p=1, N=2, t=3, i=0)  # an extra isoelectric

        self._df_subject_info = None
        self._ls_rec()

    def _ls_rec(self) -> None:
        """Find all records in the database directory
        and store them (path, metadata, etc.) in some private attributes.
        """
        super()._ls_rec()
        if (self.db_dir / "ludb.csv").is_file():
            # newly added in version 1.0.1
            self._df_subject_info = pd.read_csv(self.db_dir / "ludb.csv")
            self._df_subject_info.ID = self._df_subject_info.ID.apply(str)
            self._df_subject_info.Sex = self._df_subject_info.Sex.apply(
                lambda s: s.strip()
            )
            self._df_subject_info.Age = self._df_subject_info.Age.apply(
                lambda s: s.strip()
            )
        else:
            self._df_subject_info = pd.DataFrame(
                columns=[
                    "ID",
                    "Sex",
                    "Age",
                    "Rhythms",
                    "Electric axis of the heart",
                    "Conduction abnormalities",
                    "Extrasystolies",
                    "Hypertrophies",
                    "Cardiac pacing",
                    "Ischemia",
                    "Non-specific repolarization abnormalities",
                    "Other states",
                ]
            )

    def get_subject_id(self, rec: Union[str, int]) -> int:
        """Attach a unique subject ID for the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        int
            Subject ID associated with the record.

        """
        if isinstance(rec, int):
            rec = self[rec]
        return int(rec)

    def get_absolute_path(
        self, rec: Union[str, int], extension: Optional[str] = None
    ) -> Path:
        """Get the absolute path of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        extension : str, optional
            Extension of the file.

        Returns
        -------
        pathlib.Path
            Absolute path of the file.

        """
        if isinstance(rec, int):
            rec = self[rec]
        if extension is not None and not extension.startswith("."):
            extension = f".{extension}"
        return self.db_dir / "data" / f"{rec}{extension or ''}"

    def load_ann(
        self,
        rec: Union[str, int],
        leads: Optional[Union[str, int, Sequence[Union[str, int]]]] = None,
        metadata: bool = False,
    ) -> dict:
        """Load the wave delineation, along with metadata if specified.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        leads : str or int or List[str] or List[int], optional
            The leads of the wave delineation to be loaded.
        metadata : bool, default False
            If True, metadata will be loaded from corresponding head file.

        Returns
        -------
        ann_dict : dict
            The wave delineation annotations.

        """
        if isinstance(rec, int):
            rec = self[rec]
        ann_dict = CFG()
        rec_fp = str(self.get_absolute_path(rec))

        # wave delineation annotations
        _leads = self._normalize_leads(leads)
        _ann_ext = [
            f"{ld.lower()}" for ld in _leads
        ]  # for Version 1.0.0, it is f"{l.lower()}"
        ann_dict["waves"] = CFG({ld: [] for ld in _leads})
        for ld, ext in zip(_leads, _ann_ext):
            ann = wfdb.rdann(rec_fp, extension=ext)
            df_lead_ann = pd.DataFrame()
            symbols = np.array(ann.symbol)
            peak_inds = np.where(np.isin(symbols, ["p", "N", "t"]))[0]
            df_lead_ann["peak"] = ann.sample[peak_inds]
            df_lead_ann["onset"] = np.nan
            df_lead_ann["offset"] = np.nan
            for i, row in df_lead_ann.iterrows():
                peak_idx = peak_inds[i]
                if peak_idx == 0:
                    df_lead_ann.loc[i, "onset"] = row["peak"]
                    if symbols[peak_idx + 1] == ")":
                        df_lead_ann.loc[i, "offset"] = ann.sample[peak_idx + 1]
                    else:
                        df_lead_ann.loc[i, "offset"] = row["peak"]
                elif peak_idx == len(symbols) - 1:
                    df_lead_ann.loc[i, "offset"] = row["peak"]
                    if symbols[peak_idx - 1] == "(":
                        df_lead_ann.loc[i, "onset"] = ann.sample[peak_idx - 1]
                    else:
                        df_lead_ann.loc[i, "onset"] = row["peak"]
                else:
                    if symbols[peak_idx - 1] == "(":
                        df_lead_ann.loc[i, "onset"] = ann.sample[peak_idx - 1]
                    else:
                        df_lead_ann.loc[i, "onset"] = row["peak"]
                    if symbols[peak_idx + 1] == ")":
                        df_lead_ann.loc[i, "offset"] = ann.sample[peak_idx + 1]
                    else:
                        df_lead_ann.loc[i, "offset"] = row["peak"]
            # df_lead_ann["onset"] = ann.sample[np.where(symbols=="(")[0]]
            # df_lead_ann["offset"] = ann.sample[np.where(symbols==")")[0]]

            df_lead_ann["duration"] = (
                df_lead_ann["offset"] - df_lead_ann["onset"]
            ) * self.spacing

            df_lead_ann.index = symbols[peak_inds]

            for c in ["peak", "onset", "offset"]:
                df_lead_ann[c] = df_lead_ann[c].values.astype(int)

            for _, row in df_lead_ann.iterrows():
                w = ECGWaveForm(
                    name=self._symbol_to_wavename[row.name],
                    onset=int(row.onset),
                    offset=int(row.offset),
                    peak=int(row.peak),
                    duration=row.duration,
                )
                ann_dict["waves"][ld].append(w)

        if metadata:
            header_dict = self._load_header(rec)
            ann_dict.update(header_dict)

        return ann_dict

    def load_diagnoses(self, rec: Union[str, int]) -> List[str]:
        """Load diagnoses of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        diagnoses : List[str]
            List of diagnoses of the record.

        """
        diagnoses = self._load_header(rec)["diagnoses"]
        return diagnoses

    def load_masks(
        self,
        rec: Union[str, int],
        leads: Optional[Union[str, int, Sequence[Union[str, int]]]] = None,
        mask_format: str = "channel_first",
        class_map: Optional[Dict[str, int]] = None,
    ) -> np.ndarray:
        """Load the wave delineation in the form of masks.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        leads : str or int or List[str] or List[int], optional
            The leads of the wave delineation to be loaded.
        mask_format : str, default "channel_first"
            Format of the mask,
            "channel_last" (alias "lead_last"), or
            "channel_first" (alias "lead_first").
        class_map : dict, optional
            Custom class map.
            If not set, `self.class_map` will be used.

        Returns
        -------
        masks : numpy.ndarray
            The masks corresponding to the wave delineation annotations
            of the record.

        """
        if isinstance(rec, int):
            rec = self[rec]
        _class_map = CFG(class_map) if class_map is not None else self.class_map
        data = self.load_data(rec, leads=leads, data_format="channel_first")
        masks = np.full_like(data, fill_value=_class_map.i, dtype=int)
        waves = self.load_ann(rec, leads=leads, metadata=False)["waves"]
        for idx, (l, l_w) in enumerate(waves.items()):
            for w in l_w:
                masks[idx, w.onset : w.offset] = _class_map[
                    self._wavename_to_symbol[w.name]
                ]
        if mask_format.lower() not in [
            "channel_first",
            "lead_first",
        ]:
            masks = masks.T
        return masks

    def from_masks(
        self,
        masks: np.ndarray,
        mask_format: str = "channel_first",
        leads: Optional[Union[str, int, Sequence[Union[str, int]]]] = None,
        class_map: Optional[Dict[str, int]] = None,
        fs: Optional[Real] = None,
    ) -> Dict[str, List[ECGWaveForm]]:
        """Convert masks into lists of waveforms.

        Parameters
        ----------
        masks : numpy.ndarray
            Wave delineation in the form of masks,
            of shape ``(n_leads, seq_len)`` or ``(seq_len,)``.
        mask_format : str, default "channel_first"
            Format of the mask, used only when ``masks.ndim = 2``.
            One of "channel_last" (alias "lead_last"), or
            "channel_first" (alias "lead_first").
        leads : str or int or List[str] or List[int], optional
            The leads of the wave delineation to be loaded.
        class_map : dict, optional
            Custom class map.
            If not set, `self.class_map` will be used.
        fs : numbers.Real, optional
            Sampling frequency of the signal corresponding to the `masks`,
            If is None, `self.fs` will be used,
            to compute `duration` of the ECG waveforms.

        Returns
        -------
        waves : dict
            The wave delineation annotations of the record.
            Each item value of the dict is a list
            containing the :class:`ECGWaveForm` corr. to the lead (item key).

        """
        if masks.ndim == 1:
            _masks = masks[np.newaxis, ...]
        elif masks.ndim == 2:
            if mask_format.lower() not in [
                "channel_first",
                "lead_first",
            ]:
                _masks = masks.T
            else:
                _masks = masks.copy()
        else:
            raise ValueError(
                f"masks should be of dim 1 or 2, but got a {masks.ndim}d array"
            )

        if leads is not None:
            _leads = self._normalize_leads(leads)
        else:
            _leads = [f"lead_{idx+1}" for idx in range(_masks.shape[0])]
        assert len(_leads) == _masks.shape[0]

        _class_map = CFG(class_map) if class_map is not None else self.class_map

        _fs = fs if fs is not None else self.fs

        waves = CFG({lead_name: [] for lead_name in _leads})
        for channel_idx, lead_name in enumerate(_leads):
            current_mask = _masks[channel_idx, ...]
            for wave_symbol, wave_number in _class_map.items():
                if wave_symbol not in [
                    "p",
                    "N",
                    "t",
                ]:
                    continue
                wave_name = self._symbol_to_wavename[wave_symbol]
                current_wave_inds = np.where(current_mask == wave_number)[0]
                if len(current_wave_inds) == 0:
                    continue
                np.where(np.diff(current_wave_inds) > 1)
                split_inds = np.where(np.diff(current_wave_inds) > 1)[0].tolist()
                split_inds = sorted(split_inds + [i + 1 for i in split_inds])
                split_inds = [0] + split_inds + [len(current_wave_inds) - 1]
                for i in range(len(split_inds) // 2):
                    itv_start = current_wave_inds[split_inds[2 * i]]
                    itv_end = current_wave_inds[split_inds[2 * i + 1]] + 1
                    w = ECGWaveForm(
                        name=wave_name,
                        onset=itv_start,
                        offset=itv_end,
                        peak=np.nan,
                        duration=1000 * (itv_end - itv_start) / _fs,  # ms
                    )
                    waves[lead_name].append(w)
            waves[lead_name].sort(key=lambda w: w.onset)
        return waves

    def _load_header(self, rec: Union[str, int]) -> dict:
        """Load header data of a record into a dict.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        header_dict : dict
            The header data of the record.

        """
        header_dict = CFG({})
        rec_fp = str(self.get_absolute_path(rec))
        header_reader = wfdb.rdheader(rec_fp)
        header_dict["units"] = header_reader.units
        header_dict["baseline"] = header_reader.baseline
        header_dict["adc_gain"] = header_reader.adc_gain
        header_dict["record_fmt"] = header_reader.fmt
        try:
            header_dict["age"] = int(
                [line for line in header_reader.comments if "<age>" in line][0].split(
                    ": "
                )[-1]
            )
        except Exception:
            header_dict["age"] = np.nan
        try:
            header_dict["sex"] = [
                line for line in header_reader.comments if "<sex>" in line
            ][0].split(": ")[-1]
        except Exception:
            header_dict["sex"] = ""
        d_start = [
            idx
            for idx, line in enumerate(header_reader.comments)
            if "<diagnoses>" in line
        ][0] + 1
        header_dict["diagnoses"] = header_reader.comments[d_start:]
        return header_dict

    def load_subject_info(
        self, rec: Union[str, int], fields: Optional[Union[str, Sequence[str]]] = None
    ) -> Union[dict, str]:
        """Load subject info of a record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        fields : str or Sequence[str], optional
            Field(s) of the subject info of record.
            If not specified, all fields of the subject info will be loaded.

        Returns
        -------
        info : dict or str
            Subject info of the given fields of the record.

        """
        if isinstance(rec, int):
            rec = self[rec]
        row = self._df_subject_info[self._df_subject_info.ID == rec]
        if row.empty:
            return {}
        row = row.iloc[0]
        info = row.to_dict()
        if fields is not None:
            if isinstance(fields, str):
                assert fields in self._df_subject_info.columns, f"No field `{fields}`"
                info = info[fields]
            else:
                assert set(fields).issubset(
                    set(self._df_subject_info.columns)
                ), f"No field(s) {set(fields).difference(set(self._df_subject_info.columns))}"
                info = {k: v for k, v in info.items() if k in fields}
        return info

    def plot(
        self,
        rec: Union[str, int],
        data: Optional[np.ndarray] = None,
        ticks_granularity: int = 0,
        leads: Optional[Union[str, int, Sequence[Union[str, int]]]] = None,
        same_range: bool = False,
        waves: Optional[ECGWaveForm] = None,
        **kwargs: Any,
    ) -> None:
        """
        Plot the signals of a record or external signals (units in μV),
        with metadata (fs, labels, tranche, etc.),
        possibly also along with wave delineations.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        data : numpy.ndarray, optional
            12-lead ECG signal to plot.
            If is not None, data of `rec` will not be used.
            This is useful when plotting filtered data
        ticks_granularity : int, default 0
            Granularity to plot axis ticks, the higher the more ticks.
            0 (no ticks) --> 1 (major ticks) --> 2 (major + minor ticks)
        leads : str or int or List[str] or List[int], optional
            The leads of the ECG signal to plot.
        same_range : bool, default False
            If True, all leads are forced to have the same y range.
        waves : ECGWaveForm, optional
            The waves (p waves, t waves, qrs complexes) to plot.
        kwargs : dict, optional
            Additional keyword arguments to pass to :func:`matplotlib.pyplot.plot`.

        TODO
        ----
        1. Slice too long records, and plot separately for each segment.
        2. Plot waves using :func:`matplotlib.pyplot.axvspan`.

        NOTE
        ----
        `Locator` of ``plt`` has default `MAXTICKS` of 1000.
        If not modifying this number, at most 40 seconds of signal could be plotted once.

        Contributors: Jeethan, and WEN Hao

        """
        if isinstance(rec, int):
            rec = self[rec]
        if "plt" not in dir():
            import matplotlib.pyplot as plt

            plt.MultipleLocator.MAXTICKS = 3000

        if data is not None:
            assert leads is not None, "`leads` must be specified when `data` is given"
            data = np.atleast_2d(data)
            _leads = self._normalize_leads(leads)
            _lead_indices = [self.all_leads.index(ld) for ld in _leads]
            assert len(_leads) == data.shape[0], "number of leads must match data"
            units = self._auto_infer_units(data)
            self.logger.info(f"input data is auto detected to have units in {units}")
            if units.lower() == "mv":
                _data = 1000 * data
            else:
                _data = data
        else:
            _leads = self._normalize_leads(leads)
            _lead_indices = [self.all_leads.index(ld) for ld in _leads]
            _data = self.load_data(rec, data_format="channel_first", units="μV")[
                _lead_indices
            ]

        if same_range:
            y_ranges = np.ones((_data.shape[0],)) * np.max(np.abs(_data)) + 100
        else:
            y_ranges = np.max(np.abs(_data), axis=1) + 100

        if data is None and waves is None:
            waves = self.load_ann(rec, leads=_leads)["waves"]

        pwaves = {ld: [] for ld in _leads}
        qrs = {ld: [] for ld in _leads}
        twaves = {ld: [] for ld in _leads}
        if waves is not None:
            for l_idx, l_w in waves.items():
                for wv in l_w:
                    itv = [wv.onset, wv.offset]
                    if wv.name == self._symbol_to_wavename["p"]:
                        pwaves[l_idx].append(itv)
                    elif wv.name == self._symbol_to_wavename["N"]:
                        qrs[l_idx].append(itv)
                    elif wv.name == self._symbol_to_wavename["t"]:
                        twaves[l_idx].append(itv)

        palette = {
            "pwaves": "green",
            "qrs": "red",
            "twaves": "yellow",
        }
        plot_alpha = 0.4

        diagnoses = self.load_diagnoses(rec)

        nb_leads = len(_leads)

        seg_len = self.fs * 25  # 25 seconds
        nb_segs = _data.shape[1] // seg_len

        t = np.arange(_data.shape[1]) / self.fs
        duration = len(t) / self.fs
        fig_sz_w = int(round(4.8 * duration))
        fig_sz_h = 6 * y_ranges / 1500
        fig, axes = plt.subplots(
            nb_leads, 1, sharex=True, figsize=(fig_sz_w, np.sum(fig_sz_h))
        )
        if nb_leads == 1:
            axes = [axes]
        for idx in range(nb_leads):
            lead_name = self.all_leads[_lead_indices[idx]]
            axes[idx].plot(
                t,
                _data[idx],
                color="black",
                linewidth="2.0",
                label=f"lead - {lead_name}",
            )
            axes[idx].axhline(y=0, linestyle="-", linewidth="1.0", color="red")
            # NOTE that `Locator` has default `MAXTICKS` equal to 1000
            if ticks_granularity >= 1:
                axes[idx].xaxis.set_major_locator(plt.MultipleLocator(0.2))
                axes[idx].yaxis.set_major_locator(plt.MultipleLocator(500))
                axes[idx].grid(
                    which="major", linestyle="-", linewidth="0.5", color="red"
                )
            if ticks_granularity >= 2:
                axes[idx].xaxis.set_minor_locator(plt.MultipleLocator(0.04))
                axes[idx].yaxis.set_minor_locator(plt.MultipleLocator(100))
                axes[idx].grid(
                    which="minor", linestyle=":", linewidth="0.5", color="black"
                )
            # add extra info. to legend
            # https://stackoverflow.com/questions/16826711/is-it-possible-to-add-a-string-as-a-legend-item-in-matplotlib
            for d in diagnoses:
                axes[idx].plot([], [], " ", label=d)
            for w in ["pwaves", "qrs", "twaves"]:
                for itv in eval(f"{w}['{lead_name}']"):
                    axes[idx].axvspan(
                        itv[0] / self.fs,
                        itv[1] / self.fs,
                        color=palette[w],
                        alpha=plot_alpha,
                    )
            axes[idx].legend(loc="upper left")
            axes[idx].set_xlim(t[0], t[-1])
            axes[idx].set_ylim(-y_ranges[idx], y_ranges[idx])
            axes[idx].set_xlabel("Time [s]")
            axes[idx].set_ylabel("Voltage [μV]")
        plt.subplots_adjust(hspace=0.2)
        if kwargs.get("save_path", None):
            plt.savefig(kwargs["save_path"], dpi=200, bbox_inches="tight")
        else:
            plt.show()

    @property
    def database_info(self) -> DataBaseInfo:
        return _LUDB_INFO


__TOLERANCE = 150  # ms
__WaveNames = ["pwave", "qrs", "twave"]


def compute_metrics(
    truth_masks: Sequence[np.ndarray],
    pred_masks: Sequence[np.ndarray],
    class_map: Dict[str, int],
    fs: Real,
    mask_format: str = "channel_first",
) -> Dict[str, Dict[str, float]]:
    """Compute metrics for the wave delineation task.

    Compute metrics
    (sensitivity, precision, f1_score, mean error
    and standard deviation of the mean errors)
    for multiple evaluations

    Parameters
    ----------
    truth_masks : Sequence[numpy.ndarray]
        A sequence of ground truth masks,
        each of which can also hold multiple masks
        from different samples (differ by record or by lead).
    pred_masks : Sequence[numpy.ndarray]
        Predictions corresponding to `truth_masks`
    class_map : Dict[str, int]
        Class map, mapping names to waves to numbers from 0 to n_classes-1,
        the keys should contain "pwave", "qrs", "twave".
    fs : numbers.Real
        Sampling frequency of the signal corresponding to the masks,
        used to compute the duration of each waveform,
        hence the error and standard deviations of errors.
    mask_format : str, default "channel_first"
        Format of the mask, one of the following:
        "channel_last" (alias "lead_last"), or
        "channel_first" (alias "lead_first").

    Returns
    -------
    scorings : dict
        A dictionary containing the scorings of onsets and offsets
        of pwaves, qrs complexes, twaves.
        Each scoring is a dict consisting of the following metrics:
        sensitivity, precision, f1_score, mean_error, standard_deviation.

    """
    assert len(truth_masks) == len(pred_masks)
    truth_waveforms, pred_waveforms = [], []
    # compute for each element
    for tm, pm in zip(truth_masks, pred_masks):
        n_masks = (
            tm.shape[0]
            if mask_format.lower() in ["channel_first", "lead_first"]
            else tm.shape[1]
        )

        new_t = masks_to_waveforms(tm, class_map, fs, mask_format)
        new_t = [
            new_t[f"lead_{idx+1}"] for idx in range(n_masks)
        ]  # list of list of `ECGWaveForm`s
        truth_waveforms += new_t

        new_p = masks_to_waveforms(pm, class_map, fs, mask_format)
        new_p = [
            new_p[f"lead_{idx+1}"] for idx in range(n_masks)
        ]  # list of list of `ECGWaveForm`s
        pred_waveforms += new_p

    scorings = compute_metrics_waveform(truth_waveforms, pred_waveforms, fs)

    return scorings


def compute_metrics_waveform(
    truth_waveforms: Sequence[Sequence[ECGWaveForm]],
    pred_waveforms: Sequence[Sequence[ECGWaveForm]],
    fs: Real,
) -> Dict[str, Dict[str, float]]:
    """
    Compute the sensitivity, precision, f1_score, mean error
    and standard deviation of the mean errors,
    of evaluations on a multiple samples (differ by records, or leads).

    Parameters
    ----------
    truth_waveforms : Sequence[Sequence[ECGWaveForm]]
        The ground truth. Each element is a sequence of
        :class:`ECGWaveForm` from the same sample.
    pred_waveforms : Sequence[Sequence[ECGWaveForm]]
        The predictions corresponding to `truth_waveforms`.
        Each element is a sequence of :class:`ECGWaveForm`
        from the same sample.
    fs : numbers.Real
        Sampling frequency of the signal corresponding to the waveforms,
        used to compute the duration of each waveform,
        hence the error and standard deviations of errors.

    Returns
    -------
    scorings : dict
        A dictionary containing the scorings of onsets and offsets
        of pwaves, qrs complexes, twaves.
        Each scoring is a dict consisting of the following metrics:
        sensitivity, precision, f1_score, mean_error, standard_deviation.

    """
    truth_positive = CFG(
        {
            f"{wave}_{term}": 0
            for wave in [
                "pwave",
                "qrs",
                "twave",
            ]
            for term in ["onset", "offset"]
        }
    )
    false_positive = CFG(
        {
            f"{wave}_{term}": 0
            for wave in [
                "pwave",
                "qrs",
                "twave",
            ]
            for term in ["onset", "offset"]
        }
    )
    false_negative = CFG(
        {
            f"{wave}_{term}": 0
            for wave in [
                "pwave",
                "qrs",
                "twave",
            ]
            for term in ["onset", "offset"]
        }
    )
    errors = CFG(
        {
            f"{wave}_{term}": []
            for wave in [
                "pwave",
                "qrs",
                "twave",
            ]
            for term in ["onset", "offset"]
        }
    )
    # accumulating results
    for tw, pw in zip(truth_waveforms, pred_waveforms):
        s = _compute_metrics_waveform(tw, pw, fs)
        for wave in [
            "pwave",
            "qrs",
            "twave",
        ]:
            for term in ["onset", "offset"]:
                truth_positive[f"{wave}_{term}"] += s[f"{wave}_{term}"][
                    "truth_positive"
                ]
                false_positive[f"{wave}_{term}"] += s[f"{wave}_{term}"][
                    "false_positive"
                ]
                false_negative[f"{wave}_{term}"] += s[f"{wave}_{term}"][
                    "false_negative"
                ]
                errors[f"{wave}_{term}"] += s[f"{wave}_{term}"]["errors"]
    scorings = CFG()
    for wave in [
        "pwave",
        "qrs",
        "twave",
    ]:
        for term in ["onset", "offset"]:
            tp = truth_positive[f"{wave}_{term}"]
            fp = false_positive[f"{wave}_{term}"]
            fn = false_negative[f"{wave}_{term}"]
            err = errors[f"{wave}_{term}"]
            sensitivity = tp / (tp + fn + DEFAULTS.eps)
            precision = tp / (tp + fp + DEFAULTS.eps)
            f1_score = (
                2 * sensitivity * precision / (sensitivity + precision + DEFAULTS.eps)
            )
            mean_error = np.mean(err) * 1000 / fs
            standard_deviation = np.std(err) * 1000 / fs
            scorings[f"{wave}_{term}"] = CFG(
                sensitivity=sensitivity,
                precision=precision,
                f1_score=f1_score,
                mean_error=mean_error,
                standard_deviation=standard_deviation,
            )

    return scorings


def _compute_metrics_waveform(
    truths: Sequence[ECGWaveForm], preds: Sequence[ECGWaveForm], fs: Real
) -> Dict[str, Dict[str, float]]:
    """
    compute the sensitivity, precision, f1_score, mean error
    and standard deviation of the mean errors,
    of evaluations on a single sample (the same record, the same lead).

    Parameters
    ----------
    truths : Sequence[ECGWaveForm]
        The ground truth.
    preds : Sequence[ECGWaveForm]
        The predictions corresponding to `truths`,
    fs : numbers.Real
        Sampling frequency of the signal corresponding to the waveforms,
        used to compute the duration of each waveform,
        hence the error and standard deviations of errors.

    Returns
    -------
    scorings : dict
        A dictionary containing the scorings of onsets and offsets
        of pwaves, qrs complexes, twaves.
        Each scoring is a dict consisting of the following metrics:
        truth_positive, false_negative, false_positive, errors,
        sensitivity, precision, f1_score, mean_error, standard_deviation.

    """
    pwave_onset_truths, pwave_offset_truths, pwave_onset_preds, pwave_offset_preds = (
        [],
        [],
        [],
        [],
    )
    qrs_onset_truths, qrs_offset_truths, qrs_onset_preds, qrs_offset_preds = (
        [],
        [],
        [],
        [],
    )
    twave_onset_truths, twave_offset_truths, twave_onset_preds, twave_offset_preds = (
        [],
        [],
        [],
        [],
    )

    for item in ["truths", "preds"]:
        for w in eval(item):
            for term in ["onset", "offset"]:
                eval(f"{w.name}_{term}_{item}.append(w.{term})")

    scorings = CFG()
    for wave in [
        "pwave",
        "qrs",
        "twave",
    ]:
        for term in ["onset", "offset"]:
            (
                truth_positive,
                false_negative,
                false_positive,
                errors,
                sensitivity,
                precision,
                f1_score,
                mean_error,
                standard_deviation,
            ) = _compute_metrics_base(
                eval(f"{wave}_{term}_truths"), eval(f"{wave}_{term}_preds"), fs
            )
            scorings[f"{wave}_{term}"] = CFG(
                truth_positive=truth_positive,
                false_negative=false_negative,
                false_positive=false_positive,
                errors=errors,
                sensitivity=sensitivity,
                precision=precision,
                f1_score=f1_score,
                mean_error=mean_error,
                standard_deviation=standard_deviation,
            )
    return scorings


def _compute_metrics_base(
    truths: Sequence[Real], preds: Sequence[Real], fs: Real
) -> Tuple[int, int, int, List[float], float, float, float, float, float]:
    """The base function for computing the metrics.

    Parameters
    ----------
    truths : Sequence[Real]
        Ground truth of indices of corresponding critical points.
    preds : Sequence[Real]
        Predicted indices of corresponding critical points.
    fs : numbers.Real
        Sampling frequency of the signal corresponding to the critical points,
        used to compute the duration of each waveform,
        hence the error and standard deviations of errors.

    Returns
    -------
    tuple
        tuple of the following metrics:
        truth_positive, false_negative, false_positive, errors,
        sensitivity, precision, f1_score, mean_error, standard_deviation.

    """
    _tolerance = __TOLERANCE * fs / 1000
    _truths = np.array(truths)
    _preds = np.array(preds)
    truth_positive, false_positive, false_negative = 0, 0, 0
    errors = []
    n_included = 0
    for point in truths:
        _pred = _preds[np.where(np.abs(_preds - point) <= _tolerance)[0].tolist()]
        if len(_pred) > 0:
            truth_positive += 1
            idx = np.argmin(np.abs(_pred - point))
            errors.append(_pred[idx] - point)
        else:
            false_negative += 1
        n_included += len(_pred)

    # false_positive = len(_preds) - n_included
    false_positive = len(_preds) - truth_positive

    # print(f"""
    # truth_positive = {truth_positive}
    # false_positive = {false_positive}
    # false_negative = {false_negative}
    # """)

    # print(f"len(truths) = {len(truths)}, truth_positive + false_negative = {truth_positive + false_negative}")
    # print(f"len(preds) = {len(preds)}, truth_positive + false_positive = {truth_positive + false_positive}")

    sensitivity = truth_positive / (truth_positive + false_negative + DEFAULTS.eps)
    precision = truth_positive / (truth_positive + false_positive + DEFAULTS.eps)
    f1_score = 2 * sensitivity * precision / (sensitivity + precision + DEFAULTS.eps)
    mean_error = np.mean(errors) * 1000 / fs
    standard_deviation = np.std(errors) * 1000 / fs

    return (
        truth_positive,
        false_negative,
        false_positive,
        errors,
        sensitivity,
        precision,
        f1_score,
        mean_error,
        standard_deviation,
    )



"""
"""

from copy import deepcopy

from ...cfg import CFG

__all__ = [
    "vgg_block_basic",
    "vgg_block_mish",
    "vgg_block_swish",
    "vgg16",
    "vgg16_leadwise",
]


vgg16 = CFG()
vgg16.fs = 500
vgg16.num_convs = [2, 2, 3, 3, 3]
_base_num_filters = 12
vgg16.num_filters = [
    _base_num_filters * 4,
    _base_num_filters * 8,
    _base_num_filters * 16,
    _base_num_filters * 32,
    _base_num_filters * 32,
]
vgg16.groups = 1
vgg16.kernel_initializer = "he_normal"
vgg16.kw_initializer = {}
vgg16.activation = "relu"
vgg16.kw_activation = {}

vgg16_leadwise = deepcopy(vgg16)
vgg16_leadwise.groups = 12
_base_num_filters = 12 * 4
vgg16_leadwise.num_filters = [
    _base_num_filters * 4,
    _base_num_filters * 8,
    _base_num_filters * 16,
    _base_num_filters * 32,
    _base_num_filters * 32,
]


vgg_block_basic = CFG()
vgg_block_basic.filter_length = 15
vgg_block_basic.subsample_length = 1
vgg_block_basic.dilation = 1
vgg_block_basic.batch_norm = True
vgg_block_basic.pool_size = 3
vgg_block_basic.pool_stride = 2  # 2
vgg_block_basic.kernel_initializer = vgg16.kernel_initializer
vgg_block_basic.kw_initializer = deepcopy(vgg16.kw_initializer)
vgg_block_basic.activation = vgg16.activation
vgg_block_basic.kw_activation = deepcopy(vgg16.kw_activation)

vgg_block_mish = deepcopy(vgg_block_basic)
vgg_block_mish.activation = "mish"
del vgg_block_mish.kw_activation

vgg_block_swish = deepcopy(vgg_block_basic)
vgg_block_swish.activation = "swish"
del vgg_block_swish.kw_activation


# set default building block
vgg16.block = deepcopy(vgg_block_basic)
vgg16_leadwise.block = deepcopy(vgg_block_basic)



"""
algorithms detecting R peaks from (filtered) single-lead ECG signal

Exists algorithms from wfdb and biosppy,

TODO: algorithms compared in [1]_

NOTE:
detectors (xqrs tested) from `wfdb` better fed with signal units in mV,
if in μV, more false r peaks might be detected,
currently not tested using detectors from `biosppy`

References:
-----------
.. [1] Liu, Feifei, et al. "Performance analysis of ten common QRS detectors on different ECG application cases." Journal of healthcare engineering 2018 (2018).

"""

from numbers import Real

import numpy as np
from wfdb.processing.qrs import gqrs_detect as _gqrs_detect
from wfdb.processing.qrs import xqrs_detect as _xqrs_detect

import biosppy.signals.ecg as BSE

__all__ = [
    "xqrs_detect",
    "gqrs_detect",
    "hamilton_detect",
    "ssf_detect",
    "christov_detect",
    "engzee_detect",
    "gamboa_detect",
]


# ---------------------------------------------------------------------
# algorithms from wfdb


def xqrs_detect(sig: np.ndarray, fs: Real, **kwargs) -> np.ndarray:
    """XQRS algorithm.

    default kwargs:

        sampfrom=0, sampto='end', conf=None, learn=True, verbose=True

    """
    sig = sig.copy()  # to avoid changing the original signal
    kw = dict(sampfrom=0, sampto="end", conf=None, learn=True, verbose=False)
    kw = {k: kwargs.get(k, v) for k, v in kw.items()}
    rpeaks = _xqrs_detect(sig, fs, **kw)
    # correct R-peak locations
    (rpeaks,) = BSE.correct_rpeaks(
        signal=sig,
        rpeaks=rpeaks,
        sampling_rate=fs,
        tol=kwargs.get("correct_tol", 0.05),
    )
    return rpeaks


def gqrs_detect(sig: np.ndarray, fs: Real, **kwargs) -> np.ndarray:
    """GQRS algorithm.

    default kwargs:

        d_sig=None, adc_gain=None, adc_zero=None,
        threshold=1.0, hr=75, RRdelta=0.2, RRmin=0.28, RRmax=2.4,
        QS=0.07, QT=0.35, RTmin=0.25, RTmax=0.33,
        QRSa=750, QRSamin=130

    """
    sig = sig.copy()  # to avoid changing the original signal
    kw = dict(
        d_sig=None,
        adc_gain=None,
        adc_zero=None,
        threshold=1.0,
        hr=75,
        RRdelta=0.2,
        RRmin=0.28,
        RRmax=2.4,
        QS=0.07,
        QT=0.35,
        RTmin=0.25,
        RTmax=0.33,
        QRSa=750,
        QRSamin=130,
    )
    kw = {k: kwargs.get(k, v) for k, v in kw.items()}
    rpeaks = _gqrs_detect(sig, fs, **kw)
    # correct R-peak locations
    (rpeaks,) = BSE.correct_rpeaks(
        signal=sig,
        rpeaks=rpeaks,
        sampling_rate=fs,
        tol=kwargs.get("correct_tol", 0.05),
    )
    return rpeaks


# ---------------------------------------------------------------------
# algorithms from biosppy
def hamilton_detect(sig: np.ndarray, fs: Real, **kwargs) -> np.ndarray:
    """The default detector used by `biosppy`.

    This algorithm is based on [#ham]_.

    References
    ----------
    .. [#ham] Hamilton, Pat. "Open source ECG analysis." Computers in cardiology. IEEE, 2002.

    """
    sig = sig.copy()  # to avoid changing the original signal
    # segment
    (rpeaks,) = BSE.hamilton_segmenter(signal=sig, sampling_rate=fs)

    # correct R-peak locations
    (rpeaks,) = BSE.correct_rpeaks(
        signal=sig,
        rpeaks=rpeaks,
        sampling_rate=fs,
        tol=kwargs.get("correct_tol", 0.05),
    )
    return rpeaks


def ssf_detect(sig: np.ndarray, fs: Real, **kwargs) -> np.ndarray:
    """Slope Sum Function (SSF)

    This algorithm is originally proposed for blood pressure (BP)
    pulse detection [#ssf]_.

    References
    ----------
    .. [#ssf] Zong, W., et al. "An open-source algorithm to
              detect onset of arterial blood pressure pulses."
              Computers in Cardiology, 2003. IEEE, 2003.

    """
    sig = sig.copy()  # to avoid changing the original signal
    (rpeaks,) = BSE.ssf_segmenter(
        signal=sig,
        sampling_rate=fs,
        threshold=kwargs.get("threshold", 20),
        before=kwargs.get("before", 0.03),
        after=kwargs.get("after", 0.01),
    )
    # correct R-peak locations
    (rpeaks,) = BSE.correct_rpeaks(
        signal=sig,
        rpeaks=rpeaks,
        sampling_rate=fs,
        tol=kwargs.get("correct_tol", 0.05),
    )
    return rpeaks


def christov_detect(sig: np.ndarray, fs: Real, **kwargs) -> np.ndarray:
    """Christov detector.

    Detector proposed in [#chr]_.

    References
    ----------
    .. [#chr] Ivaylo I. Christov, "Real time electrocardiogram QRS detection
              using combined adaptive threshold", BioMedical Engineering OnLine 2004,
              vol. 3:28, 2004

    """
    sig = sig.copy()  # to avoid changing the original signal
    (rpeaks,) = BSE.christov_segmenter(signal=sig, sampling_rate=fs)
    # correct R-peak locations
    (rpeaks,) = BSE.correct_rpeaks(
        signal=sig,
        rpeaks=rpeaks,
        sampling_rate=fs,
        tol=kwargs.get("correct_tol", 0.05),
    )
    return rpeaks


def engzee_detect(sig: np.ndarray, fs: Real, **kwargs) -> np.ndarray:
    """Detector proposed by Engelse and Zeelenberg.

    This algorithm is originally proposed in [#ez1]_,
    and improved in [#ez2]_.

    References
    ----------
    .. [#ez1] W. Engelse and C. Zeelenberg, "A single scan algorithm for
              QRS detection and feature extraction", IEEE Comp. in Cardiology,
              vol. 6, pp. 37-42, 1979
    .. [#ez2] A. Lourenco, H. Silva, P. Leite, R. Lourenco and A. Fred,
              "Real Time Electrocardiogram Segmentation for Finger Based ECG Biometrics",
              BIOSIGNALS 2012, pp. 49-54, 2012

    """
    sig = sig.copy()  # to avoid changing the original signal
    (rpeaks,) = BSE.engzee_segmenter(
        signal=sig,
        sampling_rate=fs,
        threshold=kwargs.get("threshold", 0.48),
    )
    # correct R-peak locations
    (rpeaks,) = BSE.correct_rpeaks(
        signal=sig,
        rpeaks=rpeaks,
        sampling_rate=fs,
        tol=kwargs.get("correct_tol", 0.05),
    )
    return rpeaks


def gamboa_detect(sig: np.ndarray, fs: Real, **kwargs) -> np.ndarray:
    """Detector proposed by Gamboa.

    This algorithm is proposed in a PhD thesis [#gam]_.

    References
    ----------
    .. [#gam] Gamboa, Hugo. "Multi-modal behavioral biometrics based on
              HCI and electrophysiology." PhD ThesisUniversidade (2008).

    """
    sig = sig.copy()  # to avoid changing the original signal
    (rpeaks,) = BSE.gamboa_segmenter(
        signal=sig,
        sampling_rate=fs,
        tol=kwargs.get("tol", 0.48),
    )
    # correct R-peak locations
    (rpeaks,) = BSE.correct_rpeaks(
        signal=sig,
        rpeaks=rpeaks,
        sampling_rate=fs,
        tol=kwargs.get("correct_tol", 0.05),
    )
    return rpeaks



"""
The core part of the SOTA model (framework) of CPSC2020

Its key points:

1. ECG signal pre-processing: filters out the high frequency noise and baseline drift in the ECG signal
2. QRS complex detection
3. noisy heartbeat recognition: transient noise and artifacts
4. atrial fibrillation recognition: remove SPB and carefully distinguish PVC and AF beats with aberrant ventricular conduction in episodes with atrial fibrillation
5. PVC and SPB model detection: DenseNet
6. post-processing with clinical rules: a set of clinical experiences and rules including rhythm and morphological rules to suppress false positives and search for false negatives of PVC and SPB detection

"""

import math
import textwrap
from copy import deepcopy
from itertools import repeat
from typing import Optional, Sequence, Union, List

import torch
from torch import Tensor, nn

from ...cfg import CFG
from ...models._nets import Conv_Bn_Activation, DownSample
from ...utils.misc import list_sum, add_docstring, CitationMixin
from ...utils.utils_nn import (
    SizeMixin,
    compute_sequential_output_shape,
    compute_sequential_output_shape_docstring,
)


__all__ = [
    "DenseNet",
    "DenseBasicBlock",
    "DenseBottleNeck",
    "DenseMacroBlock",
    "DenseTransition",
]


if not hasattr(nn, "Dropout1d"):
    nn.Dropout1d = nn.Dropout  # added in pytorch 1.12


class DenseBasicBlock(nn.Module, SizeMixin):
    """The basic building block for DenseNet.

    This block consists of
    normalization -> activation -> convolution (-> dropout (optional)).
    The output Tensor is the concatenation of old features (input) with new features.

    Parameters
    ----------
    in_channels : int
        Number of features (channels) of the input.
    growth_rate : int
        Number of features of (channels) output from the main stream,
        further concatenated to the shortcut,
        hence making the final number of output channels grow by this value.
    filter_length : int
        Length (size) of the filter kernels.
    groups : int, default 1
        Pattern of connections between inputs and outputs.
        For more details, ref. :class:`torch.nn.Conv1d`.
    bias : bool, default False
        Whether to use bias in the convolution layers.
    dropout : float or dict, default 0.0
        Dropout rate of the new features produced from the main stream.
        If is a dict, it should contain the keys ``"p"`` and ``"type"``,
        where ``"p"`` is the dropout rate and ``"type"`` is the type of dropout,
        which can be either ``"1d"`` (:class:`torch.nn.Dropout1d`) or
        ``None`` (:class:`torch.nn.Dropout`).
    config : dict, optional
        Additional hyper-parameters, including
        activation choices, memory_efficient choices, etc.

    """

    __name__ = "DenseBasicBlock"
    __DEFAULT_CONFIG__ = CFG(
        activation="relu",
        kw_activation={"inplace": True},
        memory_efficient=False,
    )

    def __init__(
        self,
        in_channels: int,
        growth_rate: int,
        filter_length: int,
        groups: int = 1,
        bias: bool = False,
        dropout: Union[float, dict] = 0.0,
        **config,
    ) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.__growth_rate = growth_rate
        self.__kernel_size = filter_length
        self.__groups = groups
        self.config = CFG(deepcopy(self.__DEFAULT_CONFIG__))
        self.config.update(deepcopy(config))
        assert (
            in_channels % groups == 0
        ), f"`in_channels` (= `{in_channels}`) must be divisible by `groups` (= `{groups}`)"
        assert (
            growth_rate % groups == 0
        ), f"`growth_rate` (= `{growth_rate}`) must be divisible by `groups` (= `{groups}`)"

        self.bac = Conv_Bn_Activation(
            in_channels=self.__in_channels,
            out_channels=self.__growth_rate,
            kernel_size=self.__kernel_size,
            stride=1,
            dilation=1,
            groups=self.__groups,
            norm=True,
            activation=self.config.activation.lower(),
            kw_activation=self.config.kw_activation,
            bias=bias,
            ordering="bac",
        )
        if isinstance(dropout, dict):
            if dropout["type"] == "1d" and dropout["p"] > 0:
                self.dropout = nn.Dropout1d(dropout["p"])
            elif dropout["type"] is None and dropout["p"] > 0:
                self.dropout = nn.Dropout(dropout["p"])
            else:
                self.dropout = nn.Identity()
        elif dropout > 0:
            self.dropout = nn.Dropout(dropout)
        else:
            self.dropout = nn.Identity()

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass of the network.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        Returns
        -------
        output : torch.Tensor
            Output tensor,
            of shape (batch_size, n_channels, seq_len).

        """
        new_features = self.bac(input)
        new_features = self.dropout(new_features)
        if self.__groups == 1:
            output = torch.cat([input, new_features], dim=1)
        else:  # see TODO of `DenseNet`
            # input width per group
            iw_per_group = self.__in_channels // self.__groups
            # new features width per group
            nfw_per_group = self.__growth_rate // self.__groups
            output = torch.cat(
                list_sum(
                    [
                        [
                            input[:, iw_per_group * i : iw_per_group * (i + 1), :],
                            new_features[
                                :, nfw_per_group * i : nfw_per_group * (i + 1), :
                            ],
                        ]
                        for i in range(self.__groups)
                    ]
                ),
                1,
            )
        return output

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the network.

        Parameters
        ----------
        seq_len : int, optional
            Length of the 1d sequence input.
        batch_size : int, optional
            The batch size.

        Returns
        -------
        output_shape : sequence
            The output shape of the network.

        """
        out_channels = self.__in_channels + self.__growth_rate
        output_shape = (batch_size, out_channels, seq_len)
        return output_shape


class DenseBottleNeck(nn.Module, SizeMixin):
    """
    Bottleneck modification of :class:`DenseBasicBlock`,
    with an additional prefixed sequence of
    (normalization -> activation -> convolution of kernel size 1)

    Parameters
    ----------
    in_channels : int
        Number of features (channels) of the input.
    growth_rate : int
        Number of features of (channels) output from the main stream,
        further concatenated to the shortcut,
        hence making the final number of output channels grow by this value.
    bn_size : int
        Base width of intermediate layers (the bottleneck).
    filter_length : int
        Length (size) of the filter kernels of the second convolutional layer.
    groups : int, default 1
        Pattern of connections between inputs and outputs.
        For more details, ref. :class:`torch.nn.Conv1d`.
    bias : bool, default False
        Whether to use bias in the convolutional layers.
    dropout : float or dict, default 0.0
        Dropout rate of the new features produced from the main stream.
        If is a dict, it should contain the keys ``"p"`` and ``"type"``,
        where ``"p"`` is the dropout rate and ``"type"`` is the type of dropout,
        which can be either ``"1d"`` (:class:`torch.nn.Dropout1d`) or
        ``None`` (:class:`torch.nn.Dropout`).
    config : dict, optional
        Other hyper-parameters, including
        activation choices, memory_efficient choices, etc.

    """

    __name__ = "DenseBottleNeck"
    __DEFAULT_CONFIG__ = CFG(
        activation="relu",
        kw_activation={"inplace": True},
        memory_efficient=False,
    )

    def __init__(
        self,
        in_channels: int,
        growth_rate: int,
        bn_size: int,
        filter_length: int,
        groups: int = 1,
        bias: bool = False,
        dropout: Union[float, dict] = 0.0,
        **config,
    ) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.__growth_rate = growth_rate
        self.__bn_size = bn_size
        self.__kernel_size = filter_length
        self.__groups = groups
        self.config = CFG(deepcopy(self.__DEFAULT_CONFIG__))
        self.config.update(deepcopy(config))
        bottleneck_channels = self.__bn_size * self.__growth_rate

        self.neck_conv = Conv_Bn_Activation(
            in_channels=self.__in_channels,
            out_channels=bottleneck_channels,
            kernel_size=1,
            stride=1,
            dilation=1,
            groups=groups,
            norm=True,
            activation=self.config.activation.lower(),
            kw_activation=self.config.kw_activation,
            bias=bias,
            ordering="bac",
        )
        self.main_conv = Conv_Bn_Activation(
            in_channels=bottleneck_channels,
            out_channels=self.__growth_rate,
            kernel_size=self.__kernel_size,
            stride=1,
            dilation=1,
            groups=self.__groups,
            norm=True,
            activation=self.config.activation.lower(),
            kw_activation=self.config.kw_activation,
            bias=bias,
            ordering="bac",
        )
        if isinstance(dropout, dict):
            if dropout["type"] == "1d" and dropout["p"] > 0.0:
                self.dropout = nn.Dropout1d(dropout["p"])
            elif dropout["type"] is None and dropout["p"] > 0.0:
                self.dropout = nn.Dropout(dropout["p"])
            else:
                self.dropout = nn.Identity()
        elif dropout > 0.0:
            self.dropout = nn.Dropout(dropout)
        else:
            self.dropout = nn.Identity()

    def bn_function(self, input: Tensor) -> Tensor:
        """BottleNeck function.

        The "not memory_efficient" way.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        Returns
        -------
        bottleneck_output : torch.Tensor
            of shape ``(batch_size, n_channels, seq_len)``.

        """
        bottleneck_output = self.neck_conv(input)
        return bottleneck_output

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        Returns
        -------
        output : torch.Tensor
            Output tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        """
        if self.config.memory_efficient:
            raise NotImplementedError
        else:
            new_features = self.bn_function(input)
        new_features = self.main_conv(new_features)
        new_features = self.dropout(new_features)
        if self.__groups == 1:
            output = torch.cat([input, new_features], dim=1)
        else:  # see TODO of `DenseNet`
            # input width per group
            iw_per_group = self.__in_channels // self.__groups
            # new features width per group
            nfw_per_group = self.__growth_rate // self.__groups
            output = torch.cat(
                list_sum(
                    [
                        [
                            input[:, iw_per_group * i : iw_per_group * (i + 1), :],
                            new_features[
                                :, nfw_per_group * i : nfw_per_group * (i + 1), :
                            ],
                        ]
                        for i in range(self.__groups)
                    ]
                ),
                1,
            )
        return output

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the network.

        Parameters
        ----------
        seq_len : int, optional
            Length of the 1d sequence input.
        batch_size : int, optional
            The batch size.

        Returns
        -------
        output_shape : sequence
            The output shape of the network.

        """
        out_channels = self.__in_channels + self.__growth_rate
        output_shape = (batch_size, out_channels, seq_len)
        return output_shape


class DenseMacroBlock(nn.Sequential, SizeMixin):
    """Macro blocks for :class:`DenseNet`.

    Composed of a stacked sequence
    of builing blocks of similar pattern.

    Parameters
    ----------
    in_channels : int
        Number of features (channels) of the input.
    num_layers : int
        Number of building block layers.
    growth_rates: int or Sequence[int]
        Growth rate(s) for each building block layers.
        If is sequence of int, should have length equal to `num_layers`.
    bn_size : int
        Base width of intermediate layers for :class:`DenseBottleNeck`,
        not used for :class:`DenseBasicBlock`.
    filter_lengths: int or Sequence[int]
        Filter lengths(s) (kernel size(s)) for each building block layers.
        If is sequence of int, should have length equal to `num_layers`.
    groups : int, default 1
        Pattern of connections between inputs and outputs.
        For more details, ref. :class:`torch.nn.Conv1d`.
    bias : bool, default False
        Whether to use bias in convolutional layers.
    dropout : float, default 0.0,
        Dropout rate of the new features produced from the main stream
    config : dict, optional
        Other hyper-parameters, including
        extra kw for :class:`DenseBottleNeck`, and
        activation choices, memory_efficient choices, etc.

    """

    __name__ = "DenseMacroBlock"
    building_block = DenseBottleNeck

    def __init__(
        self,
        in_channels: int,
        num_layers: int,
        growth_rates: Union[Sequence[int], int],
        bn_size: int,
        filter_lengths: Union[Sequence[int], int],
        groups: int = 1,
        bias: bool = False,
        dropout: float = 0.0,
        **config,
    ) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.__num_layers = num_layers
        if isinstance(growth_rates, int):
            self.__growth_rates = list(repeat(growth_rates, num_layers))
        else:
            self.__growth_rates = list(growth_rates)
        assert len(self.__growth_rates) == self.__num_layers
        self.__bn_size = bn_size
        if isinstance(filter_lengths, int):
            self.__filter_lengths = list(repeat(filter_lengths, num_layers))
        else:
            self.__filter_lengths = list(filter_lengths)
        assert len(self.__filter_lengths) == self.__num_layers
        self.__groups = groups
        self.config = deepcopy(config)
        if self.config.get("building_block", "").lower() in [
            "basic",
            "basic_block",
        ]:
            self.building_block = DenseBasicBlock

        for idx in range(self.__num_layers):
            self.add_module(
                f"dense_building_block_{idx}",
                self.building_block(
                    in_channels=self.__in_channels + idx * self.__growth_rates[idx],
                    growth_rate=self.__growth_rates[idx],
                    bn_size=self.__bn_size,
                    filter_length=self.__filter_lengths[idx],
                    groups=self.__groups,
                    bias=bias,
                    dropout=dropout,
                    **(self.config),
                ),
            )

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the network."""
        return compute_sequential_output_shape(self, seq_len, batch_size)


class DenseTransition(nn.Sequential, SizeMixin):
    """Transition blocks between :class:`DenseMacroBlock`.

    This is a module that is
    used to perform sub-sampling,
    and compression of channels if specified.

    Parameters
    ----------
    in_channels : int
        Number of features (channels) of the input.
    compression : float, default 1.0
        Compression factor, proportion of the number of output channels
        to the number of input channels.
    subsample_length : int, default 2
        Subsampling length (size).
    groups : int, default 1
        Pattern of connections between inputs and outputs.
        For more details, ref. :class:`torch.nn.Conv1d`.
    bias : bool, default False
        Whether to use bias in convolutional layers.
    config : dict, optional
        Other parameters, including
        activation choices, subsampling mode (method), etc.

    """

    __name__ = "DenseTransition"
    __DEFAULT_CONFIG__ = CFG(
        activation="relu",
        kw_activation={"inplace": True},
        subsample_mode="avg",
    )

    def __init__(
        self,
        in_channels: int,
        compression: float = 1.0,
        subsample_length: int = 2,
        groups: int = 1,
        bias: bool = False,
        **config,
    ) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.__compression = compression
        self.__subsample_length = subsample_length
        self.__groups = groups
        assert 0 < self.__compression <= 1.0 and self.__in_channels % self.__groups == 0
        self.config = CFG(deepcopy(self.__DEFAULT_CONFIG__))
        self.config.update(deepcopy(config))

        # input width per group
        iw_per_group = self.__in_channels // self.__groups
        # new feature widths per group
        nfw_per_group = math.floor(iw_per_group * self.__compression)
        self.__out_channels = nfw_per_group * self.__groups

        self.add_module(
            "bac",
            Conv_Bn_Activation(
                in_channels=self.__in_channels,
                out_channels=self.__out_channels,
                kernel_size=1,
                stride=1,
                dilation=1,
                groups=self.__groups,
                norm=True,
                activation=self.config.activation.lower(),
                kw_activation=self.config.kw_activation,
                bias=bias,
                ordering="bac",
            ),
        )
        self.add_module(
            "down",
            DownSample(
                down_scale=self.__subsample_length,
                in_channels=self.__out_channels,
                mode=self.config.subsample_mode.lower(),
            ),
        )

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the network."""
        return compute_sequential_output_shape(self, seq_len, batch_size)


class DenseNet(nn.Sequential, SizeMixin, CitationMixin):
    """The core part of the SOTA model (framework) of CPSC2020.

    DenseNet is originally proposed in [1]_, [2]_ (journal version).
    The original implementation is available at [3]_ and [4]_. [5]_ is an
    unofficial implementation of DenseNet in PyTorch. Torchvision also provides
    an implementation of DenseNet [6]_.

    DenseNet is not only successful in image classification, but also in
    various ECG-related tasks, and is the core part of the SOTA model
    (framework) of CPSC2020.

    Parameters
    ----------
    in_channels : int
        Number of features (channels) of the input.
    config : dict
        Other hyper-parameters of the Module, ref. corresponding config file.
        Keyword arguments that must be set are as follows:

            - num_layers: sequence of int,
              number of building block layers of each dense (macro) block
            - init_num_filters: sequence of int,
              number of filters of the first convolutional layer
            - init_filter_length: sequence of int,
              filter length (kernel size) of the first convolutional layer
            - init_conv_stride: int,
              stride of the first convolutional layer
            - init_pool_size: int,
              pooling kernel size of the first pooling layer
            - init_pool_stride: int,
              pooling stride of the first pooling layer
            - growth_rates: int or sequence of int or sequence of sequences of int,
              growth rates of the building blocks,
              with granularity to the whole network, or to each dense (macro) block,
              or to each building block
            - filter_lengths: int or sequence of int or sequence of sequences of int,
              filter length(s) (kernel size(s)) of the convolutions,
              with granularity to the whole network, or to each macro block,
              or to each building block
            - subsample_lengths: int or sequence of int,
              subsampling length(s) (ratio(s)) of the transition blocks
            - compression: float,
              compression factor of the transition blocks
            - bn_size: int,
              bottleneck base width, used only when building block is :class:`DenseBottleNeck`
            - dropouts: float or dict,
              dropout ratio of each building block
            - groups: int,
              connection pattern (of channels) of the inputs and outputs
            - block: dict,
              other parameters that can be set for the building blocks

        For a full list of configurable parameters, ref. corr. config file

    NOTE
    ----
    The difference of forward output of [5]_ from others,
    however [5]_ doesnot support dropout.

    TODO
    ----
    1. For `groups` > 1, the concatenated output should be
       re-organized in the channel dimension?
    2. memory-efficient mode, i.e. storing the `new_features`
       in a shared memory instead of stacking in newly created
       :class:`~torch.Tensor` after each mini-block.

    References
    ----------
    .. [1] G. Huang, Z. Liu, L. Van Der Maaten and K. Q. Weinberger,
           "Densely Connected Convolutional Networks,"
           2017 IEEE Conference on Computer Vision and Pattern Recognition (CVPR),
           Honolulu, HI, 2017, pp. 2261-2269, doi: 10.1109/CVPR.2017.243.
    .. [2] G. Huang, Z. Liu, G. Pleiss, L. Van Der Maaten and K. Weinberger,
           "Convolutional Networks with Dense Connectivity,"
           in IEEE Transactions on Pattern Analysis and Machine Intelligence,
           doi: 10.1109/TPAMI.2019.2918284.
    .. [3] https://github.com/liuzhuang13/DenseNet/tree/master/models
    .. [4] https://github.com/gpleiss/efficient_densenet_pytorch/blob/master/models/densenet.py
    .. [5] https://github.com/bamos/densenet.pytorch/blob/master/densenet.py
    .. [6] https://github.com/pytorch/vision/blob/master/torchvision/models/densenet.py

    """

    __name__ = "DenseNet"
    __DEFAULT_CONFIG__ = CFG(
        bias=False,
        activation="relu",
        kw_activation={"inplace": True},
        kernel_initializer="he_normal",
        kw_initializer={},
        init_subsample_mode="avg",
    )

    def __init__(self, in_channels: int, **config) -> None:
        """ """
        super().__init__()
        self.__in_channels = in_channels
        self.config = CFG(deepcopy(self.__DEFAULT_CONFIG__))
        self.config.update(deepcopy(config))
        self.__num_blocks = len(self.config.num_layers)

        self.add_module(
            "init_cba",
            Conv_Bn_Activation(
                in_channels=self.__in_channels,
                out_channels=self.config.init_num_filters,
                kernel_size=self.config.init_filter_length,
                stride=1,
                dilation=1,
                groups=self.config.groups,
                norm=True,
                activation=self.config.activation.lower(),
                kw_activation=self.config.kw_activation,
                bias=self.config.bias,
            ),
        )
        self.add_module(
            "init_pool",
            DownSample(
                down_scale=self.config.init_pool_stride,
                in_channels=self.config.init_num_filters,
                kernel_size=self.config.init_pool_size,
                padding=(self.config.init_pool_size - 1) // 2,
                mode=self.config.init_subsample_mode.lower(),
            ),
        )

        if isinstance(self.config.growth_rates, int):
            self.__growth_rates = list(
                repeat(self.config.growth_rates, self.__num_blocks)
            )
        else:
            self.__growth_rates = list(self.config.growth_rates)
        assert len(self.__growth_rates) == self.__num_blocks, (
            f"`config.growth_rates` indicates {len(self.__growth_rates)} macro blocks, "
            f"while `config.num_layers` indicates {self.__num_blocks}"
        )
        if isinstance(self.config.filter_lengths, int):
            self.__filter_lengths = list(
                repeat(self.config.filter_lengths, self.__num_blocks)
            )
        else:
            self.__filter_lengths = list(self.config.filter_lengths)
            assert len(self.__filter_lengths) == self.__num_blocks, (
                f"`config.filter_lengths` indicates {len(self.__filter_lengths)} macro blocks, "
                f"while `config.num_layers` indicates {self.__num_blocks}"
            )
        if isinstance(self.config.subsample_lengths, int):
            self.__subsample_lengths = list(
                repeat(self.config.subsample_lengths, self.__num_blocks - 1)
            )
        else:
            self.__subsample_lengths = list(self.config.subsample_lengths)
            assert len(self.__subsample_lengths) == self.__num_blocks - 1, (
                f"`config.subsample_lengths` indicates {len(self.__subsample_lengths)+1} macro blocks, "
                f"while `config.num_layers` indicates {self.__num_blocks}"
            )

        macro_in_channels = self.config.init_num_filters
        for idx, macro_num_layers in enumerate(self.config.num_layers):
            dmb = DenseMacroBlock(
                in_channels=macro_in_channels,
                num_layers=macro_num_layers,
                growth_rates=self.__growth_rates[idx],
                bn_size=self.config.bn_size,
                filter_lengths=self.__filter_lengths[idx],
                groups=self.config.groups,
                bias=self.config.bias,
                dropout=self.config.dropout,
                **(self.config.block),
            )
            _, transition_in_channels, _ = dmb.compute_output_shape()
            self.add_module(f"dense_macro_block_{idx}", dmb)
            if idx < self.__num_blocks - 1:
                dt = DenseTransition(
                    in_channels=transition_in_channels,
                    compression=self.config.compression,
                    subsample_length=self.__subsample_lengths[idx],
                    groups=self.config.groups,
                    bias=self.config.bias,
                    **(self.config.transition),
                )
                _, macro_in_channels, _ = dt.compute_output_shape()
                self.add_module(f"transition_{idx}", dt)

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the network."""
        return compute_sequential_output_shape(self, seq_len, batch_size)

    @property
    def in_channels(self) -> int:
        return self.__in_channels

    @property
    def doi(self) -> List[str]:
        return list(set(self.config.get("doi", []) + ["10.1109/cvpr.2017.243"]))
